#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Numpy, a mandatory runtime dependency.
'''

# ....................{ TODO                               }....................
#FIXME: Tragically, most of this submodule no longer behaves as expected. NumPy
#2.0.0 fundamentally broke backward compatibility by removing most of the
#"numpy.__config__" attributes as well as the entire "numpy.distutils"
#subpackage required to reliably introspect NumPy parallelization.
#
#Note that, technically, there *DOES* exist a new "numpy.__config__.CONFIG"
#dictionary providing these interesting key-value pairs:
#    >>> print(numpy.__config__.CONFIG)
#    ['Build Dependencies': ['blas': ['detection method': 'pkgconfig',
#                                     'found': True,
#                                     'include directory': '/usr/include',
#                                     'lib directory': '/usr/lib64',
#                                     'name': 'cblas',
#                                     'openblas configuration': 'unknown',
#                                     'pc file directory': '/usr/lib64/pkgconfig',
#                                     'version': '3.12.0'],
#                            'lapack': ['detection method': 'pkgconfig',
#                                       'found': True,
#                                       'include directory': '/usr/include',
#                                       'lib directory': '/usr/lib64',
#                                       'name': 'lapack',
#                                       'openblas configuration': 'unknown',
#                                       'pc file directory': '/usr/lib64/pkgconfig',
#                                       'version': '3.12.0']],
#     ... ]
#
#Despite both of the above "openblas configuration" keys having a value of
#"unknown", however, this local Gentoo machine does in fact leverage OpenBLAS as
#its BLAS and LAPACK implementations. Ergo, even inspecting those keys fails to
#sufficiently detect NumPy parallelization. Since NumPy parallelization can no
#longer be reliably detected, we no longer attempt to do so. It is what it is.

#FIXME: Add detection support for NVBLAS, the Nvidia GPU-specific equivalent of
#AMD's ACML. Naturally, further research is required.

#FIXME: Consider replacing bottleneck Numpy routines with routines imported
#from the following third-party Numpy-like frameworks:
#
#* "bottleneck", providing optimized routines accepting Numpy arrays --
#  implemented in Cython and hence faster than comparible Numpy routines.
#* "numexpr", a Theano-like framework accepting Numpy arrays -- performing
#  CPU-centric parallelization of expensive array operations. Whereas Theano
#  permits such operations to be conveniently expressed in pure-Python, however,
#  numexpr inconveniently requires such operations be expressed as... wait for
#  it, raw strings. So, that sucks. Nonetheless, worth a possible look.
#* "blaze", a purported second-gen Numpy replacement. We harbour sincere doubts,
#  but everything deserves its millisecond to shine in the light. Ah. We see.
#  Blaze is considerably more heavy-weight than Numpy, and largely serves a
#  completely different marketshare: supercomputing. That's well beyond our
#  means, at the moment. Numpy it is!
#
#In short, "bottleneck" is probably the only framework listed above of interest.

#FIXME: Consider optimizing frequently used matrix and vector computations with
#Theano, a general-purpose Python mathematical optimization framework. One
#particularly compelling use case for Theano is to portably distribute
#computational work across multiple GPUs. In general, Theano can be used to
#reduce arbitrarily complex symbolic expressions expressed in pure Python to
#dynamically compiled machine code on-the-fly. Fairly amazing, all around. For
#the high-level synopsis, see:
#
#    http://deeplearning.net/software/theano/introduction.html
#FIXME: Theano and Torch (a similar framewark) appear to now be subsumed by
#TensorFlow, a Google-backed framework originally implemented in support of
#machine learning workflows at Google (e.g., DeepMind), but sufficiently
#generalized as to support a wide variety of computational needs -- like ours.

# ....................{ IMPORTS                            }....................
import numpy
from beartype.typing import (
    Dict,
    Optional,
)
from betse.util.io.log.logs import (
    log_debug,
    log_exception,
    log_warning,
)
from betse.util.os import dlls
from betse.util.os.brand import linux, macos, posix
from betse.util.path import dirs, pathnames
from betse.util.py import pys
from betse.util.py.module import pymodname, pymodule
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.iterable import itersort
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
from betse.util.type.numeric import versions
from betse.util.type.text import regexes
from betse.util.type.types import BoolOrNoneTypes, ModuleType

# ....................{ GLOBALS                            }....................
VERSION = numpy.__version__
'''
Human-readable :mod:`numpy` version string (e.g., ``1.14.5``).
'''

# ....................{ GLOBALS ~ opt_info                 }....................
# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX = None
'''
Uncompiled regular expression heuristically matching the basenames of optimized
BLAS shared libraries in the ``libraries`` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary.

This expression does *not* match the strict superset of optimized BLAS shared
libraries that are also optimized, as doing so in a cross-platform manner is
infeasible. Debian-based Linux distributions, for example, remove the prefixes
uniquely identifying the threaded variants of both ATLAS and OpenBLAS from the
basenames of their shared libraries (e.g., ``libatlas.so`` rather than
``libtatlas.so``).

Optimized BLAS shared libraries matched by this expression include:

* **AMD Core Math Library (ACML), unconditionally GPU- and CPU- optimized in
  both OpenMP-based and non-OpenMP-based variants regardless of underlying
  compiler (e.g., GNU Fortran, Open64). Note, however, that ACML does *not*
  ship with a CBLAS interface and hence is non-trivial to link Numpy against.
  While unlikely that any end users will ever do so, it nonetheless remains
  feasible and hence somewhat supported.
* **Automatically Tuned Linear Algebra Software (ATLAS),** both single- and
  multithreaded CBLAS- and Fortran-based variants for both ATLAS < 3.10 and
  ATLAS >= 3.10, which ships shared libraries under different basenames than
  ATLAS < 3.10. (Life complicates life.)
* **Intel Math Kernel Library (MKL),** unconditionally multithreaded in both
  OpenMP-based and non-OpenMP-based variants regardless of underlying compiler
  (e.g., dynamic, GCC, Intel). Note that **Intel Vector Mathematical Library
  (VML)** is intentionally ignored. Although also unconditionally
  multithreaded, VML does *not* implement the BLAS API. Numpy currently
  contains no VML-specific handling, apart from (somewhat uselessly) detecting
  VML installation on reporting system diagnostics.
* **OpenBLAS,** both single- and multithreaded 32- and 64-bit variants.

This expression is typically only required once at application startup and
hence is conditionally compiled in a just-in-time (JIT) manner by the
:func:`_is_blas_optimized_python_general` function rather than unconditionally
compiled here.
'''
# print('blas regex: ' + _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX)


# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX = None
'''
Uncompiled regular expression heuristically matching the dirnames of optimized
BLAS shared libraries in the `libraries` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''


_OPTIMIZED_BLAS_OPT_INFO_EXTRA_LINK_ARGS_MACOS = {
    # Accelerate. Although Accelerate is only conditionally multithreaded,
    # multithreading is enabled by default and hence a safe assumption.
    '-Wl,Accelerate',

    # vecLib. Similar assumptions as with Accelerate apply.
    '-Wl,vecLib',
}
'''
Set of all strings in the `extra_link_args` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary heuristically
corresponding to macOS-specific optimized BLAS shared libraries.

Unlike all other such libraries, Numpy does _not_ declare unique dictionary
globals describing macOS-specific BLAS shared libraries when linked against.
Hence, this lower-level solution.
'''

# ....................{ GLOBALS ~ linked lib               }....................
# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX = None
'''
Uncompiled regular expression heuristically matching the basenames of optimized
BLAS shared libraries dynamically linked to by Numpy.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''


# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX = None
'''
Uncompiled regular expression heuristically matching the dirnames of optimized
BLAS shared libraries dynamically linked to by Numpy.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''

# ....................{ INITIALIZERS                       }....................
# For simplicity, this function is called below on the first importation of
# this submodule rather than explicitly called by callers.
def init() -> None:
    '''
    Initialize this submodule.

    Specifically (in order):

    #. Initialize all uninitialized global variables of this submodule.
    #. If the currently installed version of Numpy was linked against an
       unoptimized BLAS implementation and is thus itself unoptimized, log a
       non-fatal warning.
    '''

    # Log this initialization.
    log_debug('Initializing NumPy...')

    # Initialize all uninitialized global variables of this submodule.
    _init_globals()

    #FIXME: Permanently disabled, as NumPy parallelization can no longer be
    #reliably detected. See "FIXME:" comment above for further details. *sigh*
    # # If Numpy linked against an unoptimized BLAS, log a non-fatal warning.
    # if not is_blas_optimized():
    #     log_warning(
    #         'Numpy unoptimized; scaling down to single-core operation. '
    #         'Consider installing an optimized multithreaded '
    #         'CBLAS implementation (e.g., OpenBLAS, ATLAS, ACML, MKL) and '
    #         'reinstalling Numpy to use this implementation.'
    #     )


def _init_globals() -> None:
    '''
    Initialize all uninitialized global variables of this submodule.
    '''

    # Permit these globals to be redefined.
    global\
        _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX,\
        _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX,\
        _OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX,\
        _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX

    # Regular expression fragment matching uniquely identifying substrings of
    # basenames of optimized BLAS shared libraries.
    blas_lib_basename_regex = r'|'.join((
        # AMD Core Math Library (ACML).
        r'acml',

        # Automatically Tuned Linear Algebra Software (ATLAS) >= 3.10.
        r'[st]?atlas',

        # Automatically Tuned Linear Algebra Software (ATLAS) < 3.10.
        r'(?:pt)?f77blas',

        # Automatically Tuned Linear Algebra Software (ATLAS) < 3.10.  Although
        # some platforms (e.g., Ubuntu) distribute the CBLAS implementation of
        # ATLAS with the ambiguous basename of "cblas" rather than "ptcblas",
        # the former may also refer to the unoptimized reference CBLAS
        # implementation and is hence ignored.
        r'ptcblas',

        # Intel Math Kernel Library (MKL). Thanks to the profusion of possible
        # library basenames, this regular expression fragment simplistically
        # assumes *ALL* library basenames prefixed by "mkl" to unconditionally
        # connote MKL. What could go wrong?
        r'mkl',

        # OpenBLAS.
        r'openblas',

        #FIXME: Research. No idea if this basename substring is even accurate.

        # BLAS-like Library Instantiation Software (BLIS). Unconditionally
        # multithreaded. Technically, Numpy has yet to add official support for
        # BLIS. Since numerous contributors nonetheless perceive BLIS to be the
        # eventual successor of BLAS *AND* since Numpy currently hosts an open
        # pull request to explicitly add BLIS support under the sensible
        # subclass name "blis_info" (see Numpy PR #7294), explicitly listing
        # BLIS here should assist in future-proofing our multithreading
        # detection.
        #r'blis',
    ))

    # Regular expression fragment matching uniquely identifying substrings of
    # dirnames of optimized BLAS shared libraries.
    blas_lib_dirname_regex = r'|'.join((
        # AMD Core Math Library (ACML).
        r'acml',

        # Automatically Tuned Linear Algebra Software (ATLAS).
        r'atlas',

        # Intel Math Kernel Library (MKL).
        r'mkl',

        # OpenBLAS.
        r'openblas',
    ))

    # Regular expression fragment matching the boundary of a dirname at which a
    # substring matching an optimized BLAS substring may either begin or end.
    dirname_boundary_regex = r'[{}_.-]'.format(dirs.SEPARATOR_REGEX)

    #FIXME: What about macOS? Are shared libraries prefixed by "lib" under that
    #platform as well?

    # Redefine this global.
    _OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX = (
        r'^lib({})(?:[_.-].*)?$'.format(blas_lib_basename_regex))

    # Redefine this global. Since Numpy has already stripped all
    # platform-specific prefixes (e.g., "lib") and suffixes (e.g., ".so") from
    # this basename, only this substring followed by an arbitrary bounded
    # suffix need be matched.
    _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX = (
        r'^({})(?:[_-].*)?$'.format(blas_lib_basename_regex))

    # Redefine these globals to the same regular expression.
    _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX = (
        r'^.*?{boundary}({blas_dirname})(?:{boundary}.*)?$'.format(
            boundary=dirname_boundary_regex,
            blas_dirname=blas_lib_dirname_regex,))
    _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX = (
        _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX)

# ....................{ TESTERS                            }....................
@func_cached
def is_blas_optimized() -> bool:
    '''
    ``True`` only if the currently installed version of Numpy is linked against
    an optimized BLAS (Basic Linear Algebra Subprograms) implementation,
    ideally but *not* necessarily parallelized across multiple processors.

    Optimized BLAS implementations are *strongly* recommended over unoptimized
    BLAS implementations. The :func:`numpy.dot` operator, which is implicitly
    optimized when Numpy is linked against a optimized BLAS implementation, is
    frequently called by BETSE in its critical path.

    Note that testing for parallelized optimized BLAS implementations, while
    more specific and hence preferable, is infeasible for common edge-cases
    (e.g., Debian-based Linux distributions). For further details, see the
    :data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX` string global.
    '''

    # For each private tester implementing a heuristic for this public test (in
    # order of decreasing generality, portability, and reliability)...
    for tester_heuristic in (
        # Detect conda-managed Numpy first, as doing so reduces to a single
        # well-defined filesystem access and hence is guaranteed to be both the
        # most optimal and portable solution.
        _is_blas_optimized_conda,
        _is_blas_optimized_opt_info_libraries,
        _is_blas_optimized_opt_info_library_dirs,
        _is_blas_optimized_opt_info_macos,
        _is_blas_optimized_posix_symlink,
    ):
        # Attempt to...
        try:
            # Log the current heuristic being attempted.
            log_debug(
                'Detecting BLAS by heuristic %s()...',
                tester_heuristic.__name__,
            )

            # Call this tester, capturing the result for subsequent handling.
            tester_result = tester_heuristic()

            # If this tester definitively identified Numpy as either
            # optimized or non-optimized...
            if tester_result is not None:
                # Log this result.
                log_debug('BLAS optimization detected: %r', tester_result)

                # Return this result.
                return tester_result
            # Else, continue to the next tester.
        # If an error occurs, log that error *WITHOUT* raising an exception.
        # Detecting Numpy optimization is non-essential and hence hardly worth
        # halting the application over.
        except Exception as exception:
            log_exception(exception)

    # Else, all heuristics failed to definitively identify Numpy to be either
    # optimized or non-optimized. For safety, assume the latter.
    return False

# ....................{ TESTERS ~ private                  }....................
def _is_blas_optimized_conda() -> BoolOrNoneTypes:
    '''
    ``True`` only if the active Python interpreter is managed by ``conda`` *or*
    ``None`` otherwise (i.e., if this interpreter is managed by any other
    means, typically a system-wide package manager).

    If the active Python interpreter is managed by ``conda``, then the current
    version of Numpy was necessarily installed from one of the following two
    prominent Anaconda channels:

    * ``anaconda``, the default proprietary Anaconda channel. In this case,
      Anaconda guarantees Numpy to be linked against Intel Math Kernel Library
      (MKL) and hence optimized.
    * ``conda-forge``, the most popular third-party open-source Anaconda
      channel. Since this and downstream applications are currently only
      available from this channel, Numpy is typically installed from this
      channel during this application's runtime. In this case, conda-forge
      guarantees Numpy to be linked against OpenBLAS and hence optimized.

    In either case, installing Numpy via ``conda`` effectively guarantees
    optimization in all sensible use cases.
    '''

    # Ultimate freedom is a working one-liner.
    #
    # Note that we intentionally return "None" rather than "False" in the event
    # that this interpreter is *NOT* managed by conda. (Returning "False" would
    # erroneously halt the detection process here.)
    return pys.is_conda() or None

# ....................{ TESTERS ~ private : opt_info       }....................
def _is_blas_optimized_opt_info_libraries() -> BoolOrNoneTypes:
    '''
    ``True`` only if the first item of the ``libraries`` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary heuristically
    corresponds to that of an optimized BLAS implementation, ``False`` if a
    non-fatal error condition arises (e.g., due this list or dictionary being
    undefined), *or* ``None`` otherwise.

    This function returns ``None`` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.

    Numpy does *not* define a public API exposing this boolean to callers.
    Numpy only defines a private API defining a medley of metadata from which
    this boolean is indirectly derivable: the :mod:`numpy.__config__`
    submodule. The :func:`numpy.misc_util.generate_config_py`
    function programmatically fabricates the contents of the
    :mod:`numpy.__config__` submodule at Numpy installation time.
    Ergo, this function introspectively inspects these contents for uniquely
    identifying metadata in a portable manner.
    '''

    # NumPy BLAS metadata if found *OR* "None" otherwise.
    blas_opt_info = _get_blas_opt_info_or_none()

    # If this metadata does *NOT* exist, silently reduce to a noop.
    if blas_opt_info is None:
        return None
    # Else, this metadata exists.

    # List of the uniquely identifying substrings of all BLAS library basenames
    # this version of Numpy is linked against in a high-level manner if any or
    # "None" otherwise.
    #
    # Note that this list is *NOT* guaranteed to exist. When this version of
    # Numpy is linked against a BLAS library in a low-level manner (e.g., via
    # "'extra_link_args': ['-Wl,-framework', '-Wl,Accelerate']" on macOS), this
    # list should *NOT* exist. In most other cases, this list should exist. To
    # avoid edge cases, this list is ignored if absent.
    blas_basename_substrs = blas_opt_info.get('libraries', None)

    # If this list is either undefined or empty, silently noop.
    if not blas_basename_substrs:
        return None
    # Else, this list is non-empty.

    # First element of this list. For simplicity, this function assumes the
    # BLAS library identified by this element currently exists. While
    # iteratively testing all listed BLAS libraries for existence would be
    # feasible, doing so is platform-specific and hence non-trivially fragile.
    blas_basename_substr = blas_basename_substrs[0]

    # If the BLAS library identified by this element is optimized, return
    # True. Since this element may be suffixed by non-identifying metadata
    # (e.g., version), a regular expression is leveraged.
    if regexes.is_match(
        text=blas_basename_substr,
        regex=_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX,
    ):
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None


def _is_blas_optimized_opt_info_library_dirs() -> BoolOrNoneTypes:
    '''
    ``True`` only if the first element of the `library_dirs` list of the
    global :data:`numpy.__config__.blas_opt_info` dictionary
    heuristically corresponds to that of an optimized BLAS implementation,
    ``False`` if a non-fatal error condition arises (e.g., due this list or
    dictionary being undefined), *or* ``None`` otherwise.

    This function returns ``None`` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.
    '''

    # NumPy BLAS metadata if found *OR* "None" otherwise.
    blas_opt_info = _get_blas_opt_info_or_none()

    # If this metadata does *NOT* exist, silently reduce to a noop.
    if blas_opt_info is None:
        return None
    # Else, this metadata exists.

    # List of the dirnames of all BLAS libraries this version of Numpy is
    # linked against in a high-level manner if any or "None" otherwise.
    #
    # Note that the "blas_opt_info" dictionary global is guaranteed to exist
    # due to the previously called _is_blas_optimized_opt_info_basename()
    # function.
    #
    # Note that this list is *NOT* guaranteed to exist. When this version of
    # Numpy is linked against a BLAS library in a low-level manner (e.g., via
    # "'extra_link_args': ['-Wl,-framework', '-Wl,Accelerate']" on macOS), this
    # list should *NOT* exist. In most other cases, this list should exist. To
    # avoid edge cases, this list is ignored if absent.
    blas_dirnames = blas_opt_info.get('library_dirs', None)

    # If this list is either undefined or empty, silently reduce to a noop.
    if not blas_dirnames:
        return None
    # Else, this list is non-empty.

    # First element of this list. For simplicity, this function assumes the
    # BLAS library identified by this element currently exists. While
    # iteratively testing all listed BLAS libraries for existence would be
    # feasible, doing so is platform-specific and hence non-trivially fragile.
    blas_dirname = blas_dirnames[0]

    # If the BLAS library identified by this element is optimized, return
    # True. Since this element is an arbitrary pathname typically containing
    # non-identifying metadata, a regular expression is leveraged.
    if regexes.is_match(
        text=blas_dirname,
        regex=_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX,
    ):
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None


def _is_blas_optimized_opt_info_macos() -> BoolOrNoneTypes:
    '''
    ``True`` only if the current platform is macOS *and* the
    ``extra_link_args`` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary both exists
    *and* heuristically corresponds to that of an optimized BLAS implementation
    specific to macOS (e.g., Accelerate, vecLib), ``False`` if a non-fatal error
    condition arises (e.g., due this list or dictionary being undefined), *or*
    ``None`` otherwise.

    This function returns ``None`` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.

    Unlike all other BLAS implementations, macOS-specific BLAS implementations
    are linked against with explicit linker flags rather than pathnames. For
    further confirmation that the
    :attr:`numpy.__config__.blas_opt_info` dictionary defines these
    flags when linked to these implementations, see:

    * https://trac.macports.org/ticket/22200
    * https://github.com/BVLC/caffe/issues/2677

    When life buys you cat food, you eat cat food.
    '''

    # If the current platform is *NOT* macOS, silently reduce to a noop.
    if not macos.is_macos():
        return None
    # Else, the current platform is macOS.

    # NumPy BLAS metadata if found *OR* "None" otherwise.
    blas_opt_info = _get_blas_opt_info_or_none()

    # If this metadata does *NOT* exist, silently reduce to a noop.
    if blas_opt_info is None:
        return None
    # Else, this metadata exists.

    # List of all implementation-specific link arguments with which Numpy
    # linked against the current BLAS implementation if any or "None".
    #
    # Note that the "blas_opt_info" dictionary global is guaranteed to exist
    # due to the previously called _is_blas_optimized_opt_info_basename()
    # function.
    blas_link_args_list = blas_opt_info.get('extra_link_args', None)

    # If no such argument exists, continue to the next heuristic. Since this
    # list is strictly optional, no errors or warnings are logged.
    if not blas_link_args_list:
        return None

    # Set of these arguments, converted from this list for efficiency.
    blas_link_args = set(blas_link_args_list)
    # logs.log_info('blas_link_args: {}'.format(blas_link_args))

    # Subset of this set specific to multithreaded BLAS implementations.
    blas_link_args_multithreaded = (
        blas_link_args & _OPTIMIZED_BLAS_OPT_INFO_EXTRA_LINK_ARGS_MACOS)

    # If this subset is nonempty, return True.
    if len(blas_link_args_multithreaded) > 0:
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None

# ....................{ TESTERS ~ private : linkage        }....................
def _is_blas_optimized_posix_symlink() -> BoolOrNoneTypes:
    '''
    ``True`` only if the current platform is POSIX-compliant and hence
    supports symbolic links *and* the first item of the ``libraries`` list of
    the global :data:`numpy.__config__.blas_ilp64_opt_info` dictionary
    is a symbolic link masquerading as either the unoptimized reference BLAS
    implementation but in fact linking to an optimized BLAS implementation.

    This function returns ``None`` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.
    '''

    # If the current platform is POSIX-incompatible and hence does *NOT*
    # support symbolic links, continue to the next heuristic.
    if not posix.is_posix():
        return None

    #FIXME: Generalize to macOS as well once the
    #libs.iter_linked_filenames() function supports macOS.

    # If the current platform is *NOT* Linux, continue to the next heuristic.
    #
    # The libs.iter_linked_filenames() function called below currently only
    # supports Linux.
    if not linux.is_linux():
        return None

    # NumPy BLAS metadata if found *OR* "None" otherwise.
    blas_opt_info = _get_blas_opt_info_or_none()

    # If this metadata does *NOT* exist, silently reduce to a noop.
    if blas_opt_info is None:
        return None
    # Else, this metadata exists.

    # First element of the list of uniquely identifying substrings of all BLAS
    # library basenames this version of Numpy is linked against.
    #
    # Note that this list is guaranteed to both exist and be non-empty due to
    # the previously called _is_blas_optimized_opt_info_basename() function.
    blas_basename_substr = blas_opt_info['libraries'][0]

    # If this element appears to be neither the reference BLAS or CBLAS
    # implementations (e.g., "blas", "cblas", "refblas", "refcblas"), continue
    # to the next heuristic.
    if not blas_basename_substr.endswith('blas'):
        return None

    # Arbitrary Numpy C extension.
    #
    # Unfortunately, the "numpy.__config__" API fails to specify the
    # absolute paths of the libraries it links against. Since there exists no
    # reliable means of reverse engineering these paths from this API, these
    # paths must be obtained by another means: specifically, by querying the
    # standard "numpy._core.multiarray" C extension installed under all supported
    # Numpy for the absolute paths of all external shared libraries to which
    # this extension links -- exactly one of which is guaranteed to be the
    # absolute path of what appears to be a reference BLAS or CBLAS
    # implementation.
    numpy_lib = get_c_extension()

    # Absolute filename of this C extension.
    numpy_lib_filename = pymodule.get_filename(module=numpy_lib)

    # For the basename and absolute filename of each shared library linked to
    # by this Numpy shared library...
    for (numpy_linked_lib_basename, numpy_linked_lib_filename) in (
        dlls.iter_linked_filenames(numpy_lib_filename)):
        # Basename excluding all suffixing filetypes of this library.
        numpy_linked_lib_rootname = pathnames.get_pathname_sans_filetypes(
            numpy_linked_lib_basename)
        # logs.log_info('rootname: %s; basename: %s; filename: %s', numpy_linked_lib_rootname, numpy_linked_lib_basename, numpy_linked_lib_filename)

        # If this appears to be neither the BLAS nor CBLAS reference library,
        # continue to the next library.
        if not numpy_linked_lib_rootname.endswith('blas'):
            continue
        # Else, this is either the BLAS or CBLAS reference library.

        # Absolute filename of the target library to which this library links
        # if this library is a symbolic link *OR* of this library as is
        # otherwise (i.e., if this is a library rather than symbolic link).
        numpy_linked_lib_target_filename = pathnames.canonicalize(
            numpy_linked_lib_filename)
        # logs.log_info('target filename: %s', numpy_linked_lib_target_filename)

        # If either the basename or dirname of this path corresponds to that of
        # an optimized BLAS library, return True.
        if regexes.is_match(
            text=pathnames.get_basename(numpy_linked_lib_target_filename),
            regex=_OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX,
        ) or regexes.is_match(
            text=pathnames.get_dirname(numpy_linked_lib_target_filename),
            regex=_OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX,
        ):
            return True

        # Else, Numpy links against an unoptimized BLAS implementation. Halt!
        break

    # Else, instruct our caller to continue to the next heuristic.
    return None

# ....................{ GETTERS                            }....................
@func_cached
def get_c_extension() -> ModuleType:
    '''
    Arbitrary Numpy-specific submodule guaranteed to be implemented as a C
    extension.

    Application startup typically tests platform-specific libraries linked
    against this C extension to attempt to dynamically detect whether this
    version of Numpy is multithreaded or not.

    :func:`get_c_extension_name_qualified`
        Further details.
    '''

    # One-liners for the greater glory of BETSE.
    return pymodname.import_module(get_c_extension_name_qualified())


@func_cached
def get_c_extension_name_qualified() -> str:
    '''
    Fully-qualified name of an arbitrary Numpy-specific submodule guaranteed to
    be implemented as a C extension.

    Specifically, this function returns either:

    * If this is Numpy >= 1.16.0, the newly unified
      :mod:`numpy._core._multiarray_umath` C extension. To quote a comment
      heading the pure-Python :mod:`numpy._core.multiarray` submodule in recent
      versions of Numpy:

          Create the ``numpy._core.multiarray`` namespace for backward
          compatibility. In v1.16 the ``multiarray`` and ``umath`` c-extension
          modules were merged into a single ``_multiarray_umath`` extension
          module. So we replicate the old namespace by importing from the
          extension module.

    * Else, the obsoleted :mod:`numpy._core.multiarray` C extension.
    '''

    # Return either...
    return (
        # If Numpy >= 1.16.0, this newly unified C extension.
        'numpy._core._multiarray_umath'
        if versions.is_greater_than_or_equal_to(VERSION, '1.16.0') else
        # Else, Numpy < 1.16.0. In this case, this obsoleted C extension.
        'numpy._core.multiarray'
    )

# ....................{ GETTERS ~ metadata                 }....................
def get_metadatas() -> tuple:
    '''
    Tuple of 2-tuples ``(metedata_name, metadata_value)``, describing all
    currently installed third-party dependencies against which Numpy was linked
    (e.g., BLAS, LAPACK).
    '''

    #FIXME: Add LAPACK linkage metadata as well.
    return (
        ('numpy (blas)', get_blas_metadata()),
    )


def get_blas_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current Numpy installation with respect
    to BLAS linkage.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # This dictionary.
    #FIXME: Permanently disabled, as NumPy parallelization can no longer be
    #reliably detected. See "FIXME:" comment above for further details. *sigh*
    # metadata = OrderedArgsDict('optimized', is_blas_optimized())
    metadata = OrderedArgsDict()

    # NumPy BLAS metadata if found *OR* "None" otherwise.
    blas_opt_info = _get_blas_opt_info_or_none()

    # If this metadata exists...
    if blas_opt_info is not None:
        # Set of all keys of the dictionary global synopsizing this metadata,
        # sorted in ascending lexicographic order for readability.
        blas_opt_info_keys = itersort.sort_ascending(
            tuple(blas_opt_info.keys()))

        # For each such key...
        for blas_opt_info_key in blas_opt_info_keys:
            # The value of this key, unconditionally converted into a string and
            # then trimmed to a reasonable string length. The values of numerous
            # keys (e.g., "libraries", "sources") commonly exceed this length,
            # hampering readability for little to no gain. Excise them all.
            metadata[blas_opt_info_key] = strs.trim(
                obj=blas_opt_info[blas_opt_info_key],
                max_len=256,
            )

    # Return this dictionary.
    return metadata

# ....................{ GETTERS ~ private                  }....................
@func_cached
def _get_blas_opt_info_or_none() -> Optional[Dict[str, object]]:
    '''
    **NumPy BLAS metadata** (i.e., dictionary mapping from human-readable
    strings to corresponding objects describing how NumPy found, linked against,
    and included headers from an external BLAS shared library at NumPy
    installation time) if such metadata exists *or* ``None`` otherwise (i.e., if
    NumPy failed to catalogue this metadata at installation time).

    Numpy does *not* define a public API exposing this metadata to callers.
    Numpy only defines a private API defining a medley of metadata, only some of
    which we particularly care about. Moreover, the attribute name of this
    metadata unreliably changes depending on various external conditions --
    including the minor version of NumPy and whether or not NumPy linked against
    a visibly 64-bit version of a BLAS library. The
    :func:`numpy.misc_util.generate_config_py` function
    programmatically fabricates the contents of the
    :mod:`numpy.__config__` submodule at Numpy installation time.
    Ergo, this function introspectively inspects these contents for uniquely
    identifying metadata in a hopefully portable manner.

    Returns
    -------
    Optional[Dict[str, object]]
        Either:

        * If NumPy BLAS metadata exists, that metadata.
        * Else, ``None``.
    '''

    # Attempt to...
    #
    # Note that most (or perhaps *ALL*) of the following logic is inherently
    # fragile. NumPy developers themselves admit that this API is private and
    # will, in all likelihood, entirely go away under a future NumPy release:
    #     This is all going to change when we change build systems, we won't be
    #     preserving the "system_info" type output.
    #
    # Future-proofing dictates that we gate *ALL* of the following logic with
    # protective exception handling guarding against mischief from NumPy devs.
    try:
        # Defer fragile imports from private third-party APIs.
        #
        # Attempt to import the "numpy.__config__" submodule first defined by
        # NumPy >= 1.23.0.
        from numpy import __config__ as numpy_config
        # If doing so fails, this version of NumPy is probably broken in various
        # ways. While we *COULD* (and possibly *SHOULD*) allow this
        # "ImportError" to unwind the call stack and terminate the active Python
        # process, we instead catch and coerce this "ImportError" into a
        # non-fatal warning below. Why? Because that's what we've always done,
        # of course. Preserving backward compatibility has tenuous value. BETSE
        # is already fragile enough. There's no upside to breaking previously
        # working behaviour, especially when we don't particularly need to.

        # NumPy install-time configuration metadata if available *OR* "None".
        # Note that this metadata should *ALWAYS* be available.
        numpy_metadata = getattr(numpy_config, 'CONFIG', None)

        # NumPy install-time BLAS configuration metadata if available *OR*
        # "None". Note that this metadata should *ALWAYS* be available.
        numpy_metadata_blas = None

        # If this metadata is defined...
        if numpy_metadata is not None:
            # NumPy install-time linkage metadata if available *OR* "None". Note
            # that this metadata should *ALWAYS* be available.
            numpy_metadata_link = numpy_metadata.get('Build Dependencies')

            # If this metadata is defined...
            if numpy_metadata_link is not None:
                # NumPy install-time BLAS linkage metadata if available *OR*
                # "None". Note that this metadata should *ALWAYS* be available.
                numpy_metadata_blas = numpy_metadata_link.get('blas')

                # If this metadata is defined, return this metadata as is.
                if numpy_metadata_blas is not None:
                    return numpy_metadata_blas
                # Else, this metadata is undefined.
            # Else, this metadata is undefined.
        # Else, this metadata is undefined.

        # Log a non-fatal warning.
        log_warning(
            (
                'NumPy %s installation misconfigured '
                '(i.e., "numpy.__config__.CONFIG[\'blas\']" '
                'subdictionary not found).'
            ),
            VERSION,
        )
    # If *ANY* of the above fragile imports failed, log a non-fatal warning.
    except ImportError:
        log_warning(
            (
                'NumPy %s unsupported '
                '(i.e., "numpy.__config__" submodule not found).'
            ),
            VERSION,
        )

    # Return "None" as a fallback, because everything else failed.
    return None



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Layer subclasses printing text over the current cell cluster.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math import mathunit
from betse.science.visual.layer.lyrabc import LayerCellsABC

# ....................{ CLASSES                            }....................
class LayerCellsIndex(LayerCellsABC):
    '''
    Layer printing the 0-based index of each cell in the cell cluster as a text
    label centered on that cell.
    '''

    # ..................{ SUPERCLASS                         }..................
    def _layer_first(self) -> None:

        # For the 0-based index and 2-tuple of X and Y coordinates of the center
        # of each cell, display this index centered at these coordinates.
        for cell_index, cell_center in enumerate(
            self._phase.cells.cell_centres):
            self._visual.axes.text(
                # Text to be displayed.
                s=cell_index,

                # X and Y coordinates to display this text at.
                x=mathunit.upscale_coordinates(cell_center[0]),
                y=mathunit.upscale_coordinates(cell_center[1]),

                # Alignment of this text at these coordinates.
                horizontalalignment='center',
                verticalalignment='center',

                # Z-order of this text with respect to other artists.
                zorder=self._zorder,
            )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.math.geometry.geopoint` submodule.
'''

# ....................{ IMPORTS                            }....................
import pytest

# ....................{ TESTS                              }....................
def test_is_left_of_vector() -> None:
    '''
    Unit test the
    :func:`betse.util.math.geometry.geopoint.is_left_of_vector` function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.geometry import geopoint

    # Head and tail points of the vector to be tested.
    vector_head_point = (0, 2)
    vector_tail_point = (0, 0)

    # Points spatially residing to the left and right of this vector to test.
    subject_point_left  = (-1, 1)
    subject_point_right = (1, -1)

    # Assert the expected spatiality of these points.
    assert geopoint.is_left_of_vector(
        subject_point=subject_point_left,
        vector_head_point=vector_head_point,
        vector_tail_point=vector_tail_point,
    )
    assert not geopoint.is_left_of_vector(
        subject_point=subject_point_right,
        vector_head_point=vector_head_point,
        vector_tail_point=vector_tail_point,
    )


def test_intersect_lines() -> None:
    '''
    Unit test the :func:`betse.util.math.geometry.geopoint.intersect_lines`
    function.
    '''

    # Defer heavyweight imports.
    import numpy
    from betse.exceptions import BetseMathLineException
    from betse.util.math.geometry import geopoint
    from numpy import array

    # Assert the expected intersection of two intersecting lines given tuples.
    assert geopoint.intersect_lines(
        line1_point1=( 1,  2),
        line1_point2=(-2, -1),
        line2_point1=(-2,  1),
        line2_point2=( 1, -2),
    ) == (-1, 0)

    # Assert the same intersection of the same intersecting lines given Numpy
    # arrays.
    intersection_array = geopoint.intersect_lines(
        line1_point1=array(( 1,  2)),
        line1_point2=array((-2, -1)),
        line2_point1=array((-2,  1)),
        line2_point2=array(( 1, -2)),
    )
    assert numpy.array_equal(intersection_array, array((-1, 0)))

    # Assert two collinear lines to have no unique intersection.
    with pytest.raises(BetseMathLineException):
        geopoint.intersect_lines(
            line1_point1=( 1,  2),
            line1_point2=(-2, -1),
            line2_point1=( 0,  1),
            line2_point2=(-1,  0),
        )

    # Assert two parallel lines to have no intersection at all.
    with pytest.raises(BetseMathLineException):
        geopoint.intersect_lines(
            line1_point1=( 1,  2),
            line1_point2=(-2, -1),
            line2_point1=( 1,  0),
            line2_point2=( 0, -1),
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level logging enumerations.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific
# modules must *ALWAYS* be able to safely import from this module at any level,
# these circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import enum, logging
from enum import IntEnum

# ....................{ ENUMS ~ level                      }....................
@enum.unique
class LogLevel(IntEnum):
    '''
    Enumeration of all possible **logging levels** (i.e., integer constants
    defined by the standard :mod:`logging` module, comparable according to the
    established semantics of the ``<`` comparator).

    This enumeration corresponds exactly to the ``--log-level`` CLI option.

    Comparison
    ----------
    Enumeration members are integer constants defined in increasing order.
    Enumeration members assigned smaller integers are more inclusive (i.e.,
    correspond to logging levels that log strictly more log messages than)
    enumeration members assigned larger integers: e.g.,

        # "DEBUG" is less and hence more inclusive than "INFO".
        >>> LogLevel.DEBUG < LogLevel.INFO
        True

    Usability
    ----------
    Enumeration members are integer constants and hence implicitly usable
    wherever an integer r-value (i.e., value in the right-hand side of a
    variable assignment statement) is expected: e.g.,

        # Instruct the root logger to entertain all log requests. Neat-o, eh?
        >>> import logging
        >>> logging.getLogger('').setLevel(LogLevel.ALL)
    '''


    ALL = logging.NOTSET
    '''
    Logging level signifying all messages to be logged.

    Since the :mod:`logging` module defines no constants encapsulating the
    concept of "all" (i.e., of logging everything), this is an ad-hoc constant
    expected to be smaller than the smallest constant defined by that module.
    '''


    DEBUG = logging.DEBUG
    '''
    Logging level suitable for debugging messages.
    '''


    INFO = logging.INFO
    '''
    Logging level suitable for informational messages.
    '''


    WARNING = logging.WARNING
    '''
    Logging level suitable for warning messages.
    '''


    ERROR = logging.ERROR
    '''
    Logging level suitable for error messages.
    '''


    CRITICAL = logging.CRITICAL
    '''
    Logging level suitable for critical messages.
    '''


    NONE = logging.CRITICAL + 1024
    '''
    Logging level signifying no messages to be logged.

    Since the :mod:`logging` module defines no constants encapsulating the
    concept of "none" (i.e., of logging nothing), this is an ad-hoc constant
    expected to be larger than the largest constant defined by that module.
    '''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **tuple** (i.e., read-only sequence *always* supporting indexation
and optionally supporting named lookup) functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import (
    type_check, ClassType, SequenceTypes, StrOrNoneTypes)
from collections import namedtuple

# ....................{ MAKERS                            }....................
@type_check
def make_named_subclass(
    # Mandatory parameters.
    class_name: str,
    item_names: SequenceTypes,

    # Optional parameters.
    doc: StrOrNoneTypes = None,
) -> ClassType:
    '''
    **Named tuple subclass** (i.e., subclass of the standard :class:`tuple`
    type whose items are accessible by both numeric indexation *and* by named
    attribute lookup ala instance and class variables) dynamically synthesized
    to have the passed class name and contain only the instance variables with
    the passed names.

    This factory function is a convenience wrapper for the :class:`namedtuple`
    function, whose non-standard semantics are arguably more obfuscatory than
    helpful.

    Parameters
    ----------
    class_name : str
        Class name of this tuple type, ideally unique across all attributes of
        the module calling this function.
    item_names : SequenceTypes
        Sequence of the names of all items of this type, required to be valid
        **Python identifiers** (i.e., contain only alphanumeric characters and
        the underscore).
    doc : StrOrNoneTypes
        Class docstring to document this type with. Defaults to ``None``, in
        which case this type remains undocumented.

    Returns
    ----------
    ClassType
        Enumeration type dynamically synthesized as defined as above.

    Examples
    ----------
        >>> from betse.util.type.iterable.tuples import make_subclass_named
        >>> OneBloodStrain = make_subclass_named(
        ...     class_name='OneBloodStrain',
        ...     item_names=('summer_goes', 'paint_leaves',))
        >>> common_man = OneBloodStrain(
        ...     summer_goes='and autumn comes',
        ...     paint_leaves='with sombre fires')
        >>> common_man.summer_goes
        'and autumn comes'
    '''

    # Avoid circular import dependencies.
    from betse.util.py import pyident
    from betse.util.type.call import callers
    from betse.util.type.iterable import itertest

    # If the passed class name is syntactically invalid, raise an exception.
    pyident.die_unless_unqualified(class_name)

    # If any passed item name is *NOT* a string, raise an exception.
    itertest.die_unless_items_instance_of(iterable=item_names, cls=str)

    # Fully-qualified name of the module defining this enumeration type.
    module_name = callers.get_caller_module_name()

    # Dynamically synthesize this named tuple subclass.
    named_subclass = namedtuple(
        typename=class_name,
        field_names=item_names,
    )

    #FIXME: Add support for passing the optional "module" parameter to the
    #namedtuple() function called above *AFTER* requiring Python >= 3.6, which
    #first introduced this parameter. Until this, we do this the hard way.

    # Set the fully-qualified name of the module declaring this subclass to
    # that of the caller. By default, the namedtuple() function sets this name
    # to that of this submodule -- fairly useless, all things considered.
    named_subclass.__module__ = module_name

    # If passed a docstring, assign this subclass this docstring.
    if doc is not None:
        named_subclass.__doc__ = doc

    # Return this subclass.
    return named_subclass



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **pickling** (i.e., serialization and deserialization of arbitrarily
complex objects to and from on-disk files) facilities.

Caveats
----------
**This submodule leverages the third-party :mod:`dill` package rather than the
standard :mod:`pickle` package.** The former conforms to the API of the latter
with additional support for so-called "exotic" types required by this
application, including:

* Generators.
* Lambda expressions.
* Ranges.
* Slices.
* Numpy :class:`ndarray` subclass instances.
* Numpy :class:`ufunc` objects.
'''

# ....................{ IMPORTS                           }....................
import dill
from betse.util.io import iofiles
from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import CALLABLE_CACHED_VAR_NAME_PREFIX
from betse.util.type.obj import objtest
from betse.util.type.types import type_check

# ....................{ CONSTANTS                         }....................
# The improved pickle-ability of protocol 4 appears to be required to pickle
# C-based data structures (e.g., "scipy.spatial.KDTree").
PROTOCOL = 4
'''
Pickle protocol used by the :func:`save` and :func:`load` functions.

This protocol is the most recent pickle protocol supported by the minimum
version of Python supported by this application, satisfying the following two
competing tradeoffs:

* Compatibility with all versions of Python supported by this application.
* Maximal **pickle-ability** (i.e., the capacity to pickle objects), improving
  support for such edge cases as very large objects and edge-case object types.
'''

# ....................{ CLASSES                           }....................
#FIXME: Given the unreliability of the private "dill._dill" API, it would be
#both substantially safer *AND* simpler to directly monkey-patch the existing
#dill.Pickler.save() method with that defined below rather than attempting to
#entirely replace the "dill._dill.Pickler" class as we currently do. Food for
#contemplative thought -- especially if dill breaks backward compatibility in
#this respect again.
class BetsePickler(dill.Pickler):
    '''
    Application-specific :mod:`dill`-based custom pickler.

    This pickler augments :mod:`dill` with additional support for
    application-specific constructs, including:

    * Preventing temporary in-memory cached data from being pickled to disk,
      including all private instance variables cached by decorators defined by
      the :mod:`betse.util.type.decorator.decmemo` module (e.g.,
      :func:`property_cached`). To do so efficiently, this pickler uncaches
      *all* previously cached data from *all* objects pickled to disk. This
      data is guaranteed to be transparently re-cached on the next in-memory
      access of this data and is thus safely uncachable. While technically
      avoidable (e.g., by saving and restoring uncached instance variables into
      a local dictionary internal to the :meth:`save` method), doing so would
      incur additional space, time, and maintenance penalties. In short, the
      lazy way still remains the best way.

    See Also
    ----------
    https://github.com/uqfoundation/dill/issues/225#issuecomment-294286518
        Feature request response by GitHub user matsjoyce_ on the :mod:`dill`
        issue tracker from which this implementation was strongly inspired.
        Thanks a metric ton, matsjoyce_!

    .. _matsjoyce:
        https://github.com/matsjoyce
    '''

    # ..................{ SAVERS                            }..................
    def save(self, obj, *args, **kwargs):
        '''
        Prepare the passed object to be pickled.

        This method is recursively called ala the Visitor pattern for each
        object to be pickled reachable in the current object graph.
        '''
        # logs.log_debug('Pickling object "%s"...', obj.__class__.__name__)

        # If this object defines a dictionary mapping from the name to value of
        # each attribute defined on this object and hence is unslotted, prevent
        # cached attributes from being pickled. Slotted objects cannot have
        # attributes dynamically added or removed at runtime and hence *MUST*
        # be ignored here.
        #
        # Since this guarantees the "obj.__dict__" attribute to exist, this
        # attribute is accessed directly below rather than indirectly via the
        # vars() builtin. While feasible, the latter is mildly less efficient.
        if hasattr(obj, '__dict__'):
            # For the name of each such attribute...
            for obj_attr_name in obj.__dict__.keys():
                # If this attribute is prefixed by a substring implying this
                # attribute to be a private instance variable to which some
                # caching decorators (e.g., @property_cached) has cached the
                # value returned by a decorated callable...
                if obj_attr_name.startswith(CALLABLE_CACHED_VAR_NAME_PREFIX):
                    # Log this deletion attempt.
                    logs.log_debug(
                        'Uncaching transient data "%s.%s"...',
                        obj.__class__.__name__, obj_attr_name)

                    # Delete this variable from this object. Again, do so
                    # directly rather than via the del() built. While feasible,
                    # the latter is slightly less efficient.
                    obj.__dict__.pop(obj_attr_name)

        # Pickle this object.
        super().save(obj, *args, **kwargs)

# ....................{ LOADERS                           }....................
@type_check
def load(filename: str) -> object:
    '''
    Load (i.e., read, unpickle, deserialize) the object previously saved to the
    file with the passed path.

    This function transparently decompresses these objects into this file when
    this filename's filetype is that of a supported archive format.

    Parameters
    ----------
    filename : str
        Absolute or relative path of this file. If this filename is suffixed by
        a supported archive filetype (i.e., if the
        :func:`betse.util.path.archives.is_filetype` function returns
        ``True`` when passed this filename), this file is automatically
        decompressed as an archive of that filetype.

    Returns
    ----------
    object
        Arbitrarily complex object and all objects transitively referenced by
        this object loaded from this file.
    '''

    # Load and return all objects saved to this file, silently decompressing
    # this file if compressed.
    with iofiles.reading_bytes(filename) as unpickle_file:
        return dill.load(file=unpickle_file)

# ....................{ SAVERS                            }....................
@type_check
def save(
    *objs,
    filename: str,
    is_overwritable: bool = False
) -> None:
    '''
    Save (i.e., write, pickle, serialize) the tuple of all passed objects to
    the file with the passed path if two or more objects are passed *or* the
    single passed object if only one object is passed.

    This function transparently compresses these objects into this file when
    this filename's filetype is that of a supported archive format.

    Parameters
    ----------
    objs : tuple
        One or more arbitrarily complex object to be serialized. These objects
        and all objects transitively referenced by this object will be
        serialized to this file. If:

        * Only one object is passed, only that object will be saved.
        * Two or more objects are passed, the tuple of all such objects will be
          saved.
    filename : str
        Absolute or relative path of this file. If this filename is suffixed by
        a supported archive filetype (i.e., if the
        :func:`betse.util.path.archives.is_filetype` function returns
        ``True`` when passed this filename), this file is automatically
        compressed into an archive of that filetype.
    is_overwritable : optional[bool]
        ``True`` if overwriting this file when this file already exists *or*
        ``False`` if raising an exception when this file already exists.
        Defaults to `False` for safety.
    '''

    # If only one object is passed, save only that object rather than the
    # 1-tuple consisting only of that object.
    if len(objs) == 1:
        objs = objs[0]

    # Save these objects to this file, silently compressing this file if this
    # filename is suffixed by an archive filetype.
    with iofiles.writing_bytes(
        filename=filename, is_overwritable=is_overwritable) as pickle_file:
        dill.dump(
            objs,
            file=pickle_file,
            protocol=PROTOCOL,

            #FIXME: We may need to increase the maximum recursion depth *BEFORE*
            #calling dill.dump() above. Let's ignore this until issues arise.

            # Physically pickle the contents of all objects transitively
            # referring to globals via stack-based recursion. By default, dill
            # non-physically pickles these objects by reference. The latter
            # approach has the benefit of avoiding recursion and hence fatal
            # errors on exceeding Python's maximum recursion depth, but the
            # distinct disadvantage of preventing pickled objects referring to
            # modules whose fully-qualified names have since changed (e.g., due
            # to those modules having since been moved, renamed, or removed)
            # from being unpickled.
            #
            # In our case, the latter is of considerably more concern than the
            # former. Whereas the former is trivially solved by increasing the
            # maximum recursion depth, the latter has no sane solution. An
            # unpicklable object is simply unpicklable. Since this codebase is
            # under constant scrutiny and construction, the fully-qualified
            # names of modules frequently change.
            #
            # To safeguard backward compatibility with previously pickled
            # objects, recursive object discovery is *STRONGLY* preferred.
            recurse=True,
        )

# ....................{ INITIALIZERS                      }....................
def init() -> None:
    '''
    Initialize both this submodule *and* the :mod:`dill` package.

    Specifically, this function instructs :mod:`dill` to pickle with our
    application-specific :class:`BetsePickler` subclass.
    '''

    # Log this initialization.
    logs.log_debug('Initializing dill...')

    # Core "dill" submodule. Sadly, dill 0.2.8.0 and 0.2.8.1 but (confusingly)
    # *NOT* dill >= 0.2.8.2 fundamentally broke backward compatibility by
    # renaming this public API -- necessitating conditional logic here to
    # properly find this submodule. For further details, see:
    #     https://github.com/uqfoundation/dill/issues/268
    dill_core_submodule = None

    # For dill >= 0.2.8.2, attempt to find the new "dill._dill" submodule.
    try:
        dill_core_submodule = dill._dill
    # For dill < 0.2.8.2, fallback to find the older "dill.dill" submodule.
    except:
        dill_core_submodule = dill.dill

    # If this submodule does *NOT* contain the expected "Pickler" base class,
    # raise an exception.
    objtest.die_unless_has_class(dill_core_submodule, 'Pickler')

    # Instruct "dill" to pickle with our application-specific pickler subclass.
    dill_core_submodule.Pickler = BetsePickler



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level upscaled object subcache functionality.
'''

# ....................{ IMPORTS                            }....................
from numpy import ndarray
from betse.science.math import mathunit
from betse.science.math.cache.cacheabc import SimPhaseCacheABC
from betse.util.type.decorator.decmemo import property_cached

# ....................{ SUBCLASSES                         }....................
class SimPhaseCacheUpscaled(SimPhaseCacheABC):
    '''
    Simulation phase-specific upscaled object subcache, persisting all
    previously constructed objects whose X and Y coordinates have been upscaled
    for use with :mod:`matplotlib` for a single simulation phase.

    Design
    ----------
    All objects persisted by this subcache are required only sporadically by
    one or more isolated features (e.g., plots, animations). Hence, these
    objects are *not* suitable for ownership by the general-purpose
    :class:`Cells` class.
    '''

    # ..................{ PROPERTIES                         }..................
    @property_cached
    def extent(self) -> tuple:
        '''
        4-tuple of the upscaled minimum and maximum X and Y coordinates of the
        environmental grid for this cell cluster.

        Specifically, this tuple consists of (in order):

        #. This grid's upscaled minimum X coordinate.
        #. This grid's upscaled maximum X coordinate.
        #. This grid's upscaled minimum Y coordinate.
        #. This grid's upscaled maximum Y coordinate.
        '''

        return mathunit.upscale_coordinates_tuple(
            self._phase.cells.xmin,
            self._phase.cells.xmax,
            self._phase.cells.ymin,
            self._phase.cells.ymax,
        )

    # ..................{ PROPERTIES ~ cells : vertices      }..................
    @property_cached
    def times_cells_vertices_coords(self) -> ndarray:
        '''
        Four-dimensional Numpy array of the upscaled coordinates of all cell
        membrane vertices for this cell cluster over all time steps under the
        assumption that these vertices are deformed over these time steps.

        See Also
        ----------
        :attrs:`betse.science.sim.Simulator.cell_verts_time`
            Further details.
        '''

        return mathunit.upscale_coordinates(
            self._phase.sim.cell_verts_time)


    @property_cached
    def cells_vertices_coords(self) -> ndarray:
        '''
        Three-dimensional Numpy array of the upscaled coordinates of all cell
        membrane vertices for this cell cluster under the assumption that these
        vertices are *not* deformed at any time step and hence are fixed over
        all time steps.

        See Also
        ----------
        :attrs:`betse.science.cells.Cells.cell_verts`
            Further details.
        '''

        return mathunit.upscale_coordinates(self._phase.cells.cell_verts)

    # ..................{ PROPERTIES ~ cells : centre        }..................
    @property_cached
    def cells_centre_x(self) -> ndarray:
        '''
        One-dimensional Numpy array of the upscaled X coordinates of all cell
        centres for this cell cluster.
        '''

        return mathunit.upscale_coordinates(
            self._phase.cells.cell_centres[:, 0])


    @property_cached
    def cells_centre_y(self) -> ndarray:
        '''
        One-dimensional Numpy array of the upscaled Y coordinates of all cell
        centres for this cell cluster.
        '''

        return mathunit.upscale_coordinates(
            self._phase.cells.cell_centres[:, 1])

    # ..................{ PROPERTIES ~ grids : centre        }..................
    @property_cached
    def grids_centre_x(self) -> ndarray:
        '''
        One-dimensional Numpy array of the upscaled X coordinates of all
        environmental grid space centres for this cell cluster.
        '''

        return mathunit.upscale_coordinates(self._phase.cells.xypts[:, 0])


    @property_cached
    def grids_centre_y(self) -> ndarray:
        '''
        One-dimensional Numpy array of the upscaled X coordinates of all
        environmental grid space centres for this cell cluster.
        '''

        return mathunit.upscale_coordinates(self._phase.cells.xypts[:, 1])



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level general-purpose string facilities.
'''

#FIXME: For maintainability, shift all:
#
#* Getter methods into a new
#  "betse.util.type.text.string.strget" submodule for maintainability.
#* Removal methods into a new
#  "betse.util.type.text.string.strcut" submodule for maintainability.
#* Wrapper methods into a new
#  "betse.util.type.text.string.strwrap" submodule for maintainability.

# ....................{ IMPORTS                           }....................
import textwrap
from betse.exceptions import BetseStrException
from betse.util.type.types import (
    type_check, IntOrNoneTypes, IterableTypes, StrOrNoneTypes)
from textwrap import TextWrapper

#FIXME: Highly non-ideal. Instead:
#
#* Define a new "betse.util.type.text.string.strtrim" submodule containing
#  *ONLY* this trim() function.
#* Import the "betse.util.type.text.string.strtrim" submodule as needed in the
#  bodies of functions in the "betse.util.type.types" submodule.
#* Refactor all current calls to betse.util.type.types.trim() to call
#  betse.util.type.text.string.strtrim.trim() instead.
#* Merge the implementations of the
#  betse.util.type.text.string.strtrim.trim() and
#  betse.util.type.text.string.strs.truncate() functions. Sadly, we appear to
#  have reimplemented the former in the latter -- largely due to this
#  nomenclature confusion. *sigh*
#* Remove the betse.util.type.types.trim() function.
#* Remove the betse.util.type.text.string.strs.truncate() function.

# For convenience, permit callers to import the general-purpose trim() function
# from this submodule rather than the "types" submodule.
from betse.util.type.types import trim

if False: trim  # silence IDE warnings

# ....................{ SINGLETONS                        }....................
text_wrapper_config = TextWrapper()
'''
Singleton :class:`TextWrapper` instance with which to wrap text.

This singleton improves efficiency -- occasionally dramatically. *All* public
functions provided by the :mod:`textwrap` module implicitly instantiate
temporary :class:`TextWrapper` instances on each function call.
'''

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_if_empty(text: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) if the passed string is empty.

    Parameters
    ----------
    text : str
        String to be validated.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string is empty.
    '''

    # If this string is empty, raise an exception.
    if not text:
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'String empty.'.format()

        # Raise this exception.
        raise BetseStrException(exception_message)


@type_check
def die_unless_prefix(
    text: str, prefix: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed arguments) if the passed string is *not*
    prefixed by the passed prefix.

    Parameters
    ----------
    text : str
        String to be validated.
    prefix : str
        Prefix to test for.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string is *not* prefixed by this prefix.

    See Also
    ----------
    :func:`is_prefix`
        Further details.
    '''

    # If this string is *NOT* prefixed by this prefix, raise an exception.
    if not is_prefix(text, prefix):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = 'String "{}" not prefixed by "{}".'.format(
                text, prefix)

        # Raise this exception.
        raise BetseStrException(exception_message)

# ....................{ EXCEPTIONS ~ substr               }....................
@type_check
def die_if_substr(
    text: str, substr: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed arguments) if the passed string contains the
    passed substring.

    Parameters
    ----------
    text : str
        String to be validated.
    substr : str
        Substring to test for.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string contains this substring.

    See Also
    ----------
    :func:`is_substr`
        Further details.
    '''

    # If this string contains this substring, raise an exception.
    if is_substr(text, substr):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = (
                'String "{}" contains substring "{}".'.format(text, substr))

        # Raise this exception.
        raise BetseStrException(exception_message)


@type_check
def die_unless_substr(
    text: str, substr: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed arguments) if the passed string does *not*
    contain the passed substring.

    Parameters
    ----------
    text : str
        String to be validated.
    substr : str
        Substring to test for.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string does *not* contain this substring.

    See Also
    ----------
    :func:`is_substr`
        Further details.
    '''

    # If this string does *NOT* contain this substring, raise an exception.
    if not is_substr(text, substr):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = (
                'String "{}" contains no substring "{}".'.format(text, substr))

        # Raise this exception.
        raise BetseStrException(exception_message)

# ....................{ TESTERS                           }....................
def is_substr(text: str, substr: str) -> bool:
    '''
    ``True`` only if the passed string does *not* contain the passed substring.

    Parameters
    ----------
    text : str
        String to be tested.
    substr : str
        Substring to test for.
    '''

    # Don't look at me like that.
    return substr in text

# ....................{ TESTERS ~ case                    }....................
@type_check
def is_lowercase(text: str) -> bool:
    '''
    ``True`` only if the passed string is strictly lowercase (i.e., this string
    contains at least one alphabetic character *and* all such characters are
    lowercase).

    This function ignores all non-alphabetic characters (e.g., digits,
    punctuation) in this string.
    '''

    # O.K., that's pretty sweet.
    return text.islower()


@type_check
def is_uppercase(text: str) -> bool:
    '''
    ``True`` only if the passed string is strictly uppercase (i.e., this string
    contains at least one alphabetic character *and* all such characters are
    uppercase).

    This function ignores all non-alphabetic characters (e.g., digits,
    punctuation) in this string.
    '''

    # Ditto on the sweetness.
    return text.isupper()

# ....................{ TESTERS ~ [pre|suf]fix            }....................
def is_prefix(text: str, prefix: str) -> bool:
    '''
    ``True`` only if the passed string is prefixed by the passed prefix.

    Parameters
    ----------
    text : str
        String to be tested.
    prefix : str
        Prefix to test for.
    '''

    return text.startswith(prefix)


@type_check
def is_suffix(text: str, suffix: str) -> bool:
    '''
    ``True`` only if the passed string is suffixed by the passed prefix.

    Parameters
    ----------
    text : str
        String to be tested.
    suffix : str
        Prefix to test for.
    '''

    return text.endswith(suffix)

# ....................{ GETTERS                           }....................
@type_check
def get_substr_first_index_or_none(text: str, substr: str) -> IntOrNoneTypes:
    '''
    0-based index of the passed substring in the passed string if any *or*
    ``None`` otherwise (i.e., if this string contains no such substring).

    Parameters
    ----------
    text : str
        String to be searched.
    substr : str
        Substring to search this string for.

    Returns
    ----------
    IntOrNoneTypes
        Either:

        * If this string contains this substring, the 0-based index of this
          substring in this string.
        * Else, ``None``.
    '''

    return text.index(substr) if substr in text else None

# ....................{ GETTERS ~ prefix                  }....................
#FIXME: Generalize all of the following getters to explicitly declare that
#these getters support arbitrary substrings -- *NOT* merely single characters.
#Since the str.index() method underlying these functions already support
#arbitrary substrings, no actual changes to the implementations of these
#functions is needed.
#
#Instead, we merely need to rename:
#
#* The "char" parameter accepted by each function to "anchor".
#* The "_char" in each function name to "_substr" -- or possibly even remove
#  the "_preceding_char" entirely.
#* Revise docstrings accordingly.

@type_check
def get_prefix_preceding_char(text: str, char: str) -> str:
    '''
    **Prefix** (i.e., substring anchored at the first character) of the passed
    string preceding the first instance of the passed character in this string
    if any *or* raise an exception otherwise (i.e., if this string contains no
    such character).

    Parameters
    ----------
    text : str
        String to be searched.
    char: str
        Character to search this string for.

    Returns
    ----------
    str
        Prefix of this string preceding the first instance of this character in
        this string.

    Raises
    ----------
    ValueError
        If this string does *not* contain this character.

    See Also
    ----------
    :func:`get_prefix_or_none`
        Getter returning ``None`` if this string contains no such character.
    :func:`get_prefix_preceding_char_or_text`
        Getter returning ``text`` if this string contains no such character.
    '''

    # Return the prefix of this string preceding the first instance of this
    # character in this string if any *OR* raise a "ValueError" otherwise.
    return text[:text.index(char)]


@type_check
def get_prefix_or_none(
    # Mandatory parameters.
    text: str,
    anchor: str,

    # Optional parameters.
    is_first: bool = True,
) -> StrOrNoneTypes:
    '''
    **Prefix** (i.e., substring anchored at the first or last substring) of the
    passed string preceding the first or last instance of the passed substring
    in this string if any *or* ``None`` otherwise (i.e., if this string
    contains no such substring).

    Parameters
    ----------
    text : str
        String to be searched.
    anchor: str
        Substring to search this string for.
    is_first : bool
        Either:

        * ``True``, in which case this prefix is anchored at (i.e., directly
          preceded by) the first instance of this substring.
        * ``False``, in which case this prefix is anchored at (i.e., directly
          preceded by) the last instance of this substring.

        Defaults to ``True``.

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * If this string contains this substring, the prefix of this string
          preceding the first or last instance of this substring.
        * Else, ``None``.

    Examples
    ----------
        >>> from betse.util.type.text.string import strs
        >>> strs.get_prefix_or_none(
        ...     text='Opposition...contradiction...premonition...compromise.',
        ...     anchor='.')
        Opposition
        >>> strs.get_prefix_or_none(
        ...     text='Opposition...contradiction...premonition...compromise.',
        ...     anchor='.',
        ...     is_first=False)
        Opposition...contradiction...premonition...compromise
        >>> strs.get_prefix_or_none(
        ...     text='This is an anomaly. Disabled. What is true?',
        ...     anchor='!')
        None
    '''

    # Return either...
    return (
        # Substring of this string preceding the first or last instance of this
        # substring in this string, conditionally depending on which the caller
        # requested...
        #
        # This seeming obfuscation is intentional premature optimization
        # eliding unnecessary statements for this frequently called function.
        text[:(text.index if is_first else text.rindex)(anchor)]
        # If this string contains this substring *OR*...
        if anchor in text else
        # Nothingness.
        None
    )


@type_check
def get_prefix_preceding_char_or_text(
    text: str, char: str, **kwargs) -> str:
    '''
    **Prefix** (i.e., substring anchored at the first or last substring) of the
    passed string preceding the first or last instance of the passed substring
    in this string if any *or* this string as is otherwise (i.e., if this
    string contains no such substring).

    Parameters
    ----------
    text : str
        String to be searched.
    char: str
        Substring to search this string for.

    All remaining keyword arguments are passed to the underlying
    :func:`get_prefix_or_none` function as is.

    Returns
    ----------
    StrOrNoneTypes
        Either:

        * If this string contains this substring, the prefix of this string
          preceding the first or last instance of this substring.
        * Else, this string as is.

    Examples
    ----------
        >>> from betse.util.type.text.string import strs
        >>> strs.get_prefix_preceding_char_or_text(
        ...     text='Seven milestones... under a watching autumn eye.',
        ...     char='.')
        Seven milestones
        >>> strs.get_prefix_preceding_char_or_text(
        ...     text='Rain's falling. Hours crawling.',
        ...     char='!')
        Rain's falling. Hours crawling.
    '''

    # Prefix of this string preceding the first instance of this character in
    # this string if any *OR* "None" otherwise.
    prefix = get_prefix_or_none(text=text, anchor=char, **kwargs)

    # Return this prefix if found *OR* this string as is otherwise.
    return prefix if prefix is not None else text

# ....................{ ADDERS                            }....................
def add_prefix_unless_found(text: str, prefix: str) -> str:
    '''
    Prefix the passed string by the passed prefix unless the former is already
    prefixed by the latter.

    Parameters
    ----------
    text : str
        String to add this prefix to.
    prefix : str
        Prefix to add to this string.

    Returns
    ----------
    str
        The former prefixed by the latter.
    '''

    return text if is_prefix(text, prefix) else prefix + text


def add_suffix_unless_found(text: str, suffix: str) -> str:
    '''
    Suffix the passed string by the passed suffix unless the former is already
    suffixed by the latter.

    Parameters
    ----------
    text : str
        String to add this suffix to.
    suffix : str
        Suffix to add to this string.

    Returns
    ----------
    str
        The former suffixed by the latter.
    '''

    return text if is_suffix(text, suffix) else text + suffix

# ....................{ CASERS                            }....................
@type_check
def lowercase_char_first(text: str) -> str:
    '''
    Lowercase the first character of the passed string.
    '''

    return text[0].lower() + text[1:] if text else ''


@type_check
def uppercase_char_first(text: str) -> str:
    '''
    Uppercase the first character of the passed string.

    Whereas the standard :meth:`str.capitalize` method both uppercases the
    first character of this string *and* lowercases all remaining characters,
    this function *only* uppercases the first character. All remaining
    characters remain unmodified.
    '''

    return text[0].upper() + text[1:] if text else ''

# ....................{ QUOTERS                           }....................
#FIXME: We don't actually escape embedded double quotes yet. The reason why is
#that doing is subtly non-trivial; for safety, we don't want to re-escape
#already escaped double quotes in this string. We only want to escape
#currently unescaped double quotes in this string. To do so, we'll need to
#leverage regex-based substitution with negative lookbehind preventing existing
#substrings matching '\\"' from being subject to substitution.

def double_quote(text: str) -> str:
    '''
    Double-quote the passed string in a human-readable manner.

    Specifically (in order):

    #. All prefixing and suffixing whitespace is removed from this string.
    #. If this string is either double- or single-quoted, these quotes are
       removed.
    #. All other double quotes in this string are escaped (e.g., replacing each
       ``"`` character with ``\\"``).
    #. The resulting string is double-quoted and returned.
    '''

    # Remove all prefixing and suffixing whitespace from this string.
    text = remove_whitespace_presuffix(text)

    # If this string is already either double- or single-quoted, strip these
    # delimiting quotes for simplicity.
    if ((text[0] == '"' and text[-1] == '"') or
        (text[0] == "'" and text[-1] == "'")):
        text = text[1:-1]

    # Double quote this string.
    return '"{}"'.format(text)

# ....................{ REMOVERS ~ prefix                 }....................
@type_check
def remove_prefix(
    text: str, prefix: str, exception_message: str = None) -> str:
    '''
    Passed string with the passed prefix removed if present *or* raise an
    exception with the passed message otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    prefix : str
        Prefix to remove from this string.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Returns
    ----------
    str
        This string truncated as detailed above.
    '''

    # If this string is *NOT* prefixed by this prefix, raise this exception.
    die_unless_prefix(text, prefix, exception_message)

    # Else, return a new string with this prefix removed from this string.
    return remove_prefix_if_found(text, prefix)


@type_check
def remove_prefix_if_found(text: str, prefix: str) -> str:
    '''
    Passed string with the passed prefix removed if present *or* the passed
    string as is otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    prefix : str
        Prefix to remove from this string.

    Returns
    ----------
    str
        This string truncated as detailed above.
    '''

    return text[len(prefix):] if is_prefix(text, prefix) else text

# ....................{ REMOVERS ~ suffix                 }....................
@type_check
def remove_suffix_if_found(text: str, suffix: str) -> str:
    '''
    Passed string with the passed suffix removed if present *or* the passed
    string as is otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    suffix : str
        Suffix to remove from this string.

    Returns
    ----------
    str
        This string truncated as detailed above.
    '''

    # There exists a special case *NOT* present in remove_prefix(). If this
    # suffix is empty, "string[:-0]" is also incorrectly empty. Avoid returning
    # the empty string in this case by explicitly testing for emptiness.
    return text[:-len(suffix)] if suffix and is_suffix(text, suffix) else text


@type_check
def remove_suffix_prefixed(text: str, suffix_prefix: str) -> str:
    '''
    Passed string with *all* characters including and following the first
    instance of the passed substring if present removed *or* the passed string
    as is otherwise.

    Specifically, this functions returns:

    * If this substring is the empty string, the empty string.
    * Else if this string contains no such substrings, this string as is.
    * Else, only the prefix of this string preceding the first such substring
      in this string.

    Parameters
    ----------
    text : str
        String to be inspected.
    suffix_prefix : str
        Substring of this string to begin removing characters at.

    Returns
    ----------
    str
        This string truncated as detailed above.

    See Also
    ----------
    :func:`remove_suffix_prefixed`
        Function replacing rather than removing this suffix.
    '''

    # If:
    #
    # * This suffix prefix is non-empty, return the substring prefixing this
    #   suffix prefix in this string if any *OR* this string as is otherwise.
    #   Fortuitously, this is exactly the first value returned by the low-level
    #   str.partition() method underlying this implementation.
    # * This suffix prefix is the empty string, return the empty string. This
    #   edge case must be explicitly tested, as str.partition() raises this
    #   exception in this case: "ValueError: empty separator".
    return text.partition(suffix_prefix)[0] if suffix_prefix else ''

# ....................{ REMOVERS ~ space                  }....................
@type_check
def remove_whitespace(text: str) -> str:
    '''
    Passed string with *all* whitespace removed, including all prefixing and
    suffixing whitespace and whitespace interspersed through this string.

    Parameters
    ----------
    text : str
        String to remove all whitespace from.

    Returns
    ----------
    str
        String with all whitespace removed.

    See Also
    ----------
    https://stackoverflow.com/a/8270124/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # While a regular expression-based implementation is also feasible, the
    # current approach of splitting this string on whitespace into substrings
    # containing no whitespace and then concatenating these substrings is
    # substantially more Pythonic (and presumably efficient).
    return ''.join(text.split())


@type_check
def remove_whitespace_presuffix(text: str) -> str:
    '''
    Passed string with all prefixing and suffixing whitespace removed.
    '''

    return text.strip()


@type_check
def remove_whitespace_suffix(text: str) -> str:
    '''
    Passed string with all suffixing (but *not* prefixing) whitespace removed.
    '''

    return text.rstrip()

# ....................{ REMOVERS ~ space : newline        }....................
@type_check
def remove_newlines_suffix(text: str) -> str:
    '''
    Passed string with all suffixing (but *not* prefixing) newlines removed.
    '''

    return text.rstrip('\n')

# ....................{ REPLACERS                         }....................
@type_check
def replace_substrs(text: str, substr: str, replacement: str) -> str:
    '''
    Passed string with all instances of the passed substring replaced by the
    passed replacement substring if any *or* the passed string as is otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    substr : str
        Substring to replace all instances of in this string.
    replacement : str
        Substring to replace all instances of the passed substring.

    Returns
    ----------
    str
        This string with all instances of this substring replaced by this
        replacement substring.

    See Also
    ----------
    :func:`betse.util.type.text.regexes.replace_substrs`
        Equivalent regular expression-based replacer.
    '''

    # ...thas just how we roll.
    return text.replace(substr, replacement)


@type_check
def replace_substr_first(text: str, substr: str, replacement: str) -> str:
    '''
    Passed string with the first instance of the passed substring replaced by
    the passed replacement substring if any *or* the passed string as is
    otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    substr : str
        Substring to replace the first instance of in this string.
    replacement : str
        Substring to replace the first instance of the passed substring.

    Returns
    ----------
    str
        This string with the first instance of this substring replaced by this
        replacement substring.
    '''

    # ...thas just how we roll.
    return text.replace(substr, replacement, 1)

# ....................{ REPLACERS ~ suffix                }....................
@type_check
def replace_suffix_prefixed(
    text: str, suffix_prefix: str, replacement: str) -> str:
    '''
    Passed string with *all* characters including and following the first
    instance of the passed substring if present replaced by the passed
    replacement substring *or* this string returned as is otherwise.

    Parameters
    ----------
    text : str
        String to be inspected.
    suffix_prefix : str
        Substring of this string to begin replacing characters at.
    replacement : str
        Substring to replace the suffix of this string prefixed by this suffix
        prefix by.

    Returns
    ----------
    str
        This string with all characters including and following the first
        instance of the passed substring replaced by this replacement
        substring.

    See Also
    ----------
    :func:`remove_suffix_prefixed`
        Function removing rather than replacing this suffix.
    '''

    # 0-based index of the first instance of this suffix prefix in this string
    # if present *OR* "None" otherwise.
    suffix_prefix_first_index = get_substr_first_index_or_none(
        text=text, substr=suffix_prefix)

    # Return...
    return (
        # If this string contains this suffix prefix, the prefix of this string
        # preceding this suffix prefix appended by this replacement substring.
        text[:suffix_prefix_first_index] + replacement
        if suffix_prefix_first_index is not None else
        # Else, this string as is.
        text
    )

# ....................{ TRUNCATERS                        }....................
@type_check
def truncate(
    text: str,
    max_len: int = 80,
    suffix_prefix: StrOrNoneTypes = None,
    replacement: str = '...',
) -> str:
    '''
    Passed string truncated to the passed suffix prefix and maximum length if
    applicable *or* this string as is otherwise.

    Specifically, this function applies the following operations (in order):

    #. If a suffix prefix is passed, the suffix of the passed string prefixed
       by this prefix is replaced by the passed replacement.
    #. If the resulting string still exceeds the passed maximum length, the
       suffix of this string exceeding this length is again replaced by the
       passed replacement.
    #. The resulting string is returned.

    Parameters
    ----------
    text : str
        String to be truncated.
    max_len : int
        Maximum number of characters to truncate this string to. Defaults to
        the standard UNIX terminal line length (i.e., 80).
    suffix_prefix : StrOrNoneTypes
        Substring of this string to begin prematurely truncating characters at
        if this string contains this substring, regardless of whether this
        string exceeds this maximum length. This substring imposes a "barrier."
        Defaults to ``None``, in which case no such barrier is imposed.
    replacement : str
        Substring to replace the truncated portion of this string with.
        Defaults to an ASCII ellipses (i.e., ``...``).

    Returns
    ----------
    str
        This string truncated to this maximum length, as detailed above.
    '''

    # If passed a suffix prefix, replace the suffix of this string prefixed
    # by this prefix by this replacement.
    if suffix_prefix is not None:
        text = replace_suffix_prefixed(
            text=text, suffix_prefix=suffix_prefix, replacement=replacement)

    # If this string does *NOT* exceed this maximum, return this string as is.
    if len(text) <= max_len:
        return text
    # Else, this string exceeds this maximum. In this case...
    else:
        # Number of characters to truncate from the end of this string.
        # Dismantled, this is:
        #
        # * "len(text) - max_len", the number of characters that this string
        #   exceeds this maximum length by.
        # * "... + len(replacement)", truncating an additional number of
        #   characters equal to the length of this replacement so as to make
        #   sufficient room for this replacement at the end of this string
        #   without exceeding this maximum length.
        #
        # Note that this number is guaranteed to be non-negative (i.e., greater
        # than zero), as "len(text) > max_len" and "len(replacement) >= 0".
        truncate_chars_count = len(text) - max_len + len(replacement)

        # If more characters are to be truncated from this string than exist in
        # this string, then it can be shown by trivial algebraic equivalency
        # that "len(replacement) > max_len" (i.e., the length of the
        # replacement substring exceeds the maximum length). In this uncommon
        # edge case, return the replacement truncated to this maximum.
        if truncate_chars_count > len(text):
            return replacement[:max_len]
        # Else, fewer characters are to be truncated from this string than
        # exist in this string. This is the common case.

        # Return this string truncated to this number of characters appended by
        # this replacement substring.
        return text[:-truncate_chars_count] + replacement

# ....................{ WRAPPERS                          }....................
@type_check
def wrap_lines(lines: IterableTypes, **kwargs) -> str:
    '''
    Wrap the passed iterable of lines to the passed line width, prefixing each
    resulting wrapped line by the passed line prefix.

    See Also
    ----------
    :func:`wrap`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # Wrap it up, please.
    return wrap(strjoin.join(lines), **kwargs)


@type_check
def wrap(
    text: str,
    text_wrapper = textwrap,
    line_prefix: str = '',
    **kwargs
) -> str:
    '''
    Wrap the passed text to the passed line width, prefixing each resulting
    wrapped line by the passed line prefix.

    Parameters
    ----------
    This function accepts the following keyword arguments in addition to those
    accepted by the :func:`textwrap.wrap` function:

    * ``text_wrapper`, the object on which to call the :func:`wrap` function or
      method. For safety, this defaults to the :mod:`textwrap` module.
    * ``line_prefix``, the substring prefixing each wrapped output line.

    See Also
    ----------
    https://docs.python.org/3/library/textwrap.html
        For further details on keyword arguments.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    assert hasattr(text_wrapper, 'wrap'), (
        'Object "{}" not a text wrapper '
        '(i.e., has no wrap() callable).'.format(text_wrapper))

    # wrap() function or method to be called.
    wrap_callable = getattr(text_wrapper, 'wrap')
    assert callable(wrap_callable), (
        'Text wrapper "{}" attribute "wrap" not callable.'.format(
            text_wrapper))

    # If passed a nonempty line prefix, add appropriate keyword arguments.
    if line_prefix:
        kwargs['initial_indent'] = line_prefix
        kwargs['subsequent_indent'] = line_prefix

    # For reliability, call the wrap() function of module "textwrap" rather than
    # the singleton object "text_wrapper". The former internally instantiates a
    # new instance of class "TextWrapper" and hence resets all wrapping
    # attributes to sensible defaults, whereas the latter reuses existing such
    # attributes -- which may no longer retain sensible defaults.
    return strjoin.join_on_newline(wrap_callable(text, **kwargs))

# ....................{ WRAPPERS ~ un                     }....................
@type_check
def unwrap(text: str) -> str:
    '''
    Convert the passed possibly **multiline string** (i.e., string containing
    one or more newlines) to a **single-line string** (i.e., string containing
    no newlines).
    '''

    # Nice one, stdlib.
    return text.replace('\n', ' ')

# ....................{ (IN|DE)DENTERS                    }....................
def dedent(*texts) -> str:
    '''
    Remove all indentation shared in common by all lines of all passed strings.
    '''

    return textwrap.dedent(*texts)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`pymod`-based decorators marking tests and fixtures as
**parametrized** (i.e., accepting two or more sets of parameters passed to
these tests and fixtures).

Most of these decorators mark tests and fixtures with application-specific
keywords inspected *only* by custom :mod:`pytest` hooks defined by custom
``conftest`` plugins.
'''

# ....................{ IMPORTS                           }....................
import itertools, pytest
from betse.exceptions import BetseTestParamException
from betse.util.type.types import (
    type_check, CallableTypes, ContainerTypes, MappingType, SequenceTypes,)

# ....................{ PARAMS                            }....................
# Outer decorator accepting all parameters explicitly passed to this decorator.
@type_check
def parametrize_test_setwise(
    params: MappingType = None,
    fixtures: MappingType = None,
    ids: SequenceTypes = None,
) -> CallableTypes:
    '''
    Optionally directly parametrize all non-fixture parameters accepted by the
    decorated test callable with the first passed dictionary *and* optionally
    indirectly parametrize all fixture parameters accepted by this callable
    with the second passed dictionary.

    Fixtures
    ----------
    This decorator should *not* be directly applied to fixtures, for which the
    :func:`pytest.fixture` decorator should be applied instead as follows:

    * Pass that decorator:

      * The mandatory ``params`` keyword argument, whose value is the sequence
        of all sequences of parameters to be iteratively passed to that
        fixture.
      * The optional ``ids`` keyword argument, whose value is the sequence of
        all human-readable unique identifiers to be assigned to each such set
        of parameters (in the same order). While optional, this argument
        premains highly recommended.

    * Pass that fixture the standard ``request`` fixture, whose ``param``
      attribute supplies the set of parameters passed to the current fixture
      invocation.

    Fixture Parameters
    ----------
    For each fixture accepted by the decorated test callable, the
    :func:`pytest.fixture` decorator should be applied to that fixture as
    follows:

    * Pass that fixture the `request` builtin fixture, whose ``param``
      attribute supplies the set of parameters passed to the current fixture
      invocation.

    Avoid passing that decorator either the ``params`` or ``ids`` keyword
    argument, whose values will be indirectly supplied by this decorator to
    that fixture.

    Parameters
    ----------
    params : optional[MappingType]
        Dictionary mapping from the names of non-fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (in order), such that:

        #. The first item of each such sequence is the first value of the
           corresponding parameter to be passed to the first parametrization of
           this test.
        #. The second item of the same sequence is the second value of this
           parameter to be passed to the second parametrization of this test.
        #. And so on.

        Defaults to ``None``, in which case this test is assumed to accept no
        non-fixture parameters.
    fixtures : optional[MappingType]
        Dictionary mapping from the names of fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (in order) such that the same
        interpretation as for ``params`` holds. Defaults to ``None``, in which
        case this test is assumed to accept no fixture parameters.
    ids : optional[SequenceTypes]
        Sequence of all human-readable labels uniquely identifying each
        parametrization of this test (in the same order). Defaults to ``None``,
        in which non-human-readable labels will be automatically synthesized
        from the values of the parameters comprising these parametrizations.

    Raises
    ----------
    BetseTestParamException
        If the ``ids`` argument identifies a different number of
        parametrizations than the ``params`` and ``fixtures`` arguments
        actually passed.

    Examples
    ----------
        >>> from pytest import fixture
        >>> from betse.util.test.pytest.mark.pytparam import parametrize_test_setwise
        >>> @fixture
        ... def feathered_serpent(serpent_name: str) -> str:
        ...     if serpent_name == 'Hualpa'
        ...         return 'Amazonia'
        ...     elif serpent_name == 'Mujaji'
        ...         return 'Azania'
        >>> @parametrize_test_setwise(
        ...     params={
        ...         'western_dragon': ('Celedyr', 'Hestaby',),
        ...         'eastern_dragon': ('Masaru', 'Ryumyo',),
        ...     },
        ...     fixtures={
        ...         'feathered_serpent': ('Hualpa', 'Mujaji',),
        ...     },
        ...     ids=('bad-dragons', 'good-dragons',),
        ... )
        ... def test_great_dragons(
        ...     western_dragon: str, eastern_dragon: str, feathered_serpent: str):
        ...     assert western_dragon in ('Celedyr', 'Hestaby',)
        ...     assert eastern_dragon in ('Masaru', 'Ryumyo',)
        ...     assert feathered_serpent in ('Amazonia', 'Azania',)
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable.iterables import zip_isometric

    # Default unpassed arguments to the empty dictionary for sanity. For unknown
    # (but presumably valid) reasons, these defaults *MUST* be established here
    # rather than in the inner closure defined below. Just. Because.
    if params   is None: params   = {}
    if fixtures is None: fixtures = {}

    # Inner closure decorating the actual test callable.
    def _parametrize_test_fixtures_inner(test_callable) -> CallableTypes:
        # Validate all non-fixture parametrizations to be sequences.
        for param_values in params.values():
            if not isinstance(param_values, SequenceTypes):
                raise BetseTestParamException(
                    'Parametrized test "{}" '
                    'non-fixture parametrization {} not a sequence.'.format(
                    test_callable.__name__, param_values))

        # Validate all fixture parametrizations to be sequences.
        for fixture_values in fixtures.values():
            if not isinstance(fixture_values, SequenceTypes):
                raise BetseTestParamException(
                    'Parametrized test "{}" '
                    'fixture parametrization {} not a sequence.'.format(
                    test_callable.__name__, fixture_values))

        # Tuple of the names of all fixture parameters (in arbitrary order) if
        # any or False otherwise. Attempting to pass an empty tuple as the
        # "indirect" keyword argument to the @pytest.mark.parametrize
        # decorator raises the following non-human-readable exception:
        #
        #     /usr/lib64/python3.4/site-packages/_pytest/python.py:1952: in pytest_generate_tests
        #         argnames = func_params[0]
        #     E   IndexError: tuple index out of range
        #
        # Likewise, attempting to pass None results in even more obscure
        # non-human-readable exceptions resembling:
        #
        #     /usr/lib64/python3.4/site-packages/_pytest/python.py:869: in setmulti
        #         valtype_for_arg = valtypes[arg]
        #     E   KeyError: '{param_name}'
        #
        # ...where "{param_name}" is the name of the first parameter listed in
        # the "param_fixture_names" tuple defined below.
        #
        # py.test, you are nice; but your failure to raise human-readable
        # errors is a deplorable waste worthy of unclean roadhouse outhouses.
        fixture_names = tuple(fixtures.keys()) or False

        # Tuple of the names of both non-fixture and fixture parameters (in a
        # predictable, but arbitrary, order).
        #
        # Note that the dict.keys() method returns instances of the "dict_keys"
        # class; unlike standard sequence classes (e.g., "list", "tuple"), this
        # class does *NOT* overload the "+" operator. Aggregating multiple
        # instances of this class requires leveraging the generic
        # itertools.chain() function, accepting arbitrary iterables. (Ugh.)
        param_fixture_names = tuple(itertools.chain(
            params.keys(), fixtures.keys()))
        # param_fixture_names = ','.join(itertools.chain(
        #     params.keys(), fixtures.keys()))

        # Tuple of n-tuples of all values for each parametrization of these
        # non-fixture and parameters (in this same predictable order). See
        # parametrize_test_setwise() for further discussion.
        param_fixture_values = tuple(zip_isometric(*itertools.chain(
            params.values(), fixtures.values())))

        # If the caller identified a different number of parametrizations than
        # were actually passed, raise an exception.
        if ids is not None and len(ids) != len(param_fixture_values):
            raise BetseTestParamException(
                'Parametrized test "{}" '
                'number of parametrization identifiers {} differs from '
                'number of parametrizations {}.'.format(
                 test_callable.__name__, len(ids), len(param_fixture_values)))
        # print('\n    test: {}; param_names: {!r}; values: {!r}; fixtures: {!r}; ids: {!r}'.format(test_callable.__name__, param_fixture_names, param_fixture_values, fixture_names, ids))

        # Defer to the canonical parametrize() decorator.
        return pytest.mark.parametrize(
            argnames=param_fixture_names,
            argvalues=param_fixture_values,
            indirect=fixture_names,
            ids=ids,
        )(test_callable)

    # Return this inner closure.
    return _parametrize_test_fixtures_inner

# ....................{ PARAMS ~ setwise                  }....................
# Outer decorator accepting all parameters explicitly passed to this decorator.

#FIXME: Actually implement. See the "Examples" docstring section for the
#intended usage of this decorator; basically, it provides an alternative means
#(and arguably equally intuitive) of parametrizing tests to the above
#parametrize_test_setwise() decorator.
#FIXME: The principal disadvantage of both this and the above
#parametrize_test_setwise() decorator is that neither supports the fine-grained
#specification of specific parametrizations to be either skipped or xfailed, as
#supported by the lower-level @pytest.mark.parametrize decorator. See the
#following official py.test documentation for a useful example:
#    http://doc.pytest.org/en/latest/skipping.html#skip-xfail-with-parametrize
#While the above parametrize_test_setwise() decorator *CANNOT* (by design) be
#refactored to support such fine-grained specification, this decorator should
#be able to do so. How? By defining two new "ParametrizationSkipIf" and
#"ParametrizationXFail" classes instantiated as follows:
#
#    @parametrize_test_setwise(
#        parametrizations={
#            'bad-dragons': {
#                'western_dragon':    'Celedyr',
#                'eastern_dragon':    'Masaru',
#                'feathered_serpent': 'Hualpa',
#            },
#            'good-dragons': ParametrizationXFail(
#                parametrizations={
#                    'western_dragon':    'Hestaby',
#                    'eastern_dragon':    'Ryumyo',
#                    'feathered_serpent': 'Mujaji',
#                },
#                reason="Good dragons raise unexpected breath attack.",
#            },
#        },
#        fixture_names={'feathered_serpent',},
#    )
#    def test_great_dragons(
#        western_dragon: str, eastern_dragon: str, feathered_serpent: str):
#        assert western_dragon in ('Celedyr', 'Hestaby',)
#        assert eastern_dragon in ('Masaru', 'Ryumyo',)
#        assert feathered_serpent in ('Amazonia', 'Azania',)
#
#This decorator would then need to explicitly differentiate such objects from
#standard dictionaries and behave as expected. Quite a bit of work, sadly, but
#certainly feasible.
#FIXME: Repair documentation.

@type_check
def parametrize_test_paramwise(
    fixture_names: ContainerTypes = None,
    parametrizations: MappingType = None,
) -> CallableTypes:
    '''
    Optionally directly parametrize all non-fixture parameters accepted by the
    decorated test callable with the first passed dictionary *and* optionally
    indirectly parametrize all fixture parameters accepted by this callable
    with the second passed dictionary.

    Fixtures
    ----------
    This decorator should *not* be directly applied to fixtures, for which the
    :func:`pytest.fixture` decorator should be applied instead as follows:

    * Pass that decorator:

      * The mandatory ``params`` keyword argument, whose value is the sequence
        of all sequences of parameters to be iteratively passed to that
        fixture.
      * The optional ``ids`` keyword argument, whose value is the sequence of
        all human-readable unique identifiers to be assigned to each such set
        of parameters (in the same order). While optional, this argument
        premains highly recommended.

    * Pass that fixture the standard ``request`` builtin, whose ``param``
      attribute supplies the set of parameters passed to the current fixture
      invocation.

    Fixture Parameters
    ----------
    For each fixture accepted by the decorated test callable, the
    :func:`pytest.fixture` decorator should be applied to that fixture as
    follows:

    * Pass that fixture the standard ``request`` fixture, whose ``param``
      attribute supplies the set of parameters passed to the current fixture
      invocation.

    Avoid passing that decorator either the ``params`` or ``ids`` keyword
    argument, whose values will be indirectly supplied by this decorator to
    that fixture.

    Parameters
    ----------
    params : optional[MappingType]
        Dictionary mapping from the names of non-fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (in order), such that:

        #. The first item of each such sequence is the first value of the
           corresponding parameter to be passed to the first parametrization of
           this test.
        #. The second item of the same sequence is the second value of this
           parameter to be passed to the second parametrization of this test.
        #. And so on.

        Defaults to ``None``, in which case this test is assumed to accept no
        non-fixture parameters.
    fixtures : optional[MappingType]
        Dictionary mapping from the names of fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (in order) such that the same
        interpretation as for ``params`` holds. Defaults to ``None``, in which
        case this test is assumed to accept no fixture parameters.
    ids : optional[SequenceTypes]
        Sequence of all human-readable labels uniquely identifying each
        parametrization of this test (in the same order). Defaults to ``None``,
        in which non-human-readable labels will be automatically synthesized
        from the values of the parameters comprising each such parametrization.

    Raises
    ----------
    BetseTestParamException
        If the ``ids`` argument identifies a different number of parametrizations
        than the ``params`` and ``fixtures`` arguments actually passed.

    Examples
    ----------
        >>> from pytest import fixture
        >>> from betse.util.test.pytest.mark.pytparam import parametrize_test_setwise
        >>> @fixture
        ... def feathered_serpent(serpent_name: str) -> str:
        ...     if serpent_name == 'Hualpa'
        ...         return 'Amazonia'
        ...     elif serpent_name == 'Mujaji'
        ...         return 'Azania'
        >>> @parametrize_test_setwise(
        ...     parametrizations={
        ...         'bad-dragons': {
        ...             'western_dragon':    'Celedyr',
        ...             'eastern_dragon':    'Masaru',
        ...             'feathered_serpent': 'Hualpa',
        ...         },
        ...         'good-dragons': {
        ...             'western_dragon':    'Hestaby',
        ...             'eastern_dragon':    'Ryumyo',
        ...             'feathered_serpent': 'Mujaji',
        ...         },
        ...     },
        ...     fixture_names={'feathered_serpent',},
        ... )
        ... def test_great_dragons(
        ...     western_dragon: str, eastern_dragon: str, feathered_serpent: str):
        ...     assert western_dragon in ('Celedyr', 'Hestaby',)
        ...     assert eastern_dragon in ('Masaru', 'Ryumyo',)
        ...     assert feathered_serpent in ('Amazonia', 'Azania',)
    '''

    raise ValueError('This decorator currently unimplemented.')

# ....................{ PARAMS ~ serial                   }....................
serialize_parametrized_test = pytest.mark.serialize_parametrized_test
'''
Mark the decorated parametrized test as **serial** (i.e., parametrized such
that the success of each subsequent parameter set depends on the success of all
previous parameter sets for the current test requesting this decoration).

If this test is *not* parametrized, an exception is raised. Else, each
parameter set passed to this test is tested serially. On the first failure of
this test passed a parameter set, each subsequent call to this test passed a
subsequent parameter set will be automatically marked as an ``XFAIL`` and hence
fail *without* being run.

Implementation
----------
The majority of the black magic required by this decoration is implemented as
low-level :mod:`pytest` hooks in the top-level :mod:`betse_test.conftest`
plugin. To preserve state between parametrized calls to the same test, these
hooks dynamically add the following application-specific attributes to this
test's underlying function or method object:

* ``_betse_first_failing_param_id``, the unique identifier of the first set of
  parameter values passed to this test raising an exception for the current
  test session if any *or* ``None`` otherwise (i.e., if this test has yet to
  raise an exception for any parameters).
'''


def serialize_parametrized_fixture(fixture):
    '''
    Decorator marking the decorated parametrized fixture as **serial** (i.e.,
    parametrized such that the success of each subsequent parameter set depends
    on the success of all previous parameter sets for the current test
    requesting this decoration).

    If this fixture is *not* parametrized, an exception is raised.

    This decoration is automatically propagated to all tests requesting this
    fixture. Decorating a fixture with this decorator is functionally
    equivalent to decorating all tests requesting this fixture by the
    :mod:`serialize_parametrized_test` decorator.

    Implementation
    ----------
    Technically, :mod:`pytest` currently ignores marks decorating fixtures. To
    circumvent this omission, this decorator adds the following
    application-specific attribute to this fixture's underlying function or
    method object:

    * ``_betse_is_fixture_parametrized_serially``, ``True`` only if this
      fixture is parametrized serially. If this fixture is *not* parametrized
      serially, this attribute is usually undefined.

    See Also
    ----------
    :func:`serialize_parametrized_test`
        Further details on serial parametrization.
    '''

    # Mark this fixture as serially parametrized. Since fixtures do *NOT*
    # support marks, an application-specific attribute is injected into this
    # fixture instead.
    fixture._betse_is_fixture_parametrized_serially = True
    return fixture



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :mod:`betse.util.type.iterable.iterget` submodule.
'''

# ....................{ IMPORTS                           }....................

# ....................{ CLASSES                           }....................
class BellType(object):
    '''
    Type of each item of the iterable to be tested by the
    :func:`test_get_item_var_uniquified_str` unit test.

    Attributes
    ----------
    name : str
        Name of this item, guaranteed to be unique across all such items.
    stanza : str
        Arbitrary string associated with this item.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, name: str, stanza: str) -> None:

        self.name = name
        self.stanza = stanza

# ....................{ TESTS ~ default                   }....................
def test_get_item_var_uniquified_str() -> None:
    '''
    Unit test the
    :func:`betse.util.type.iterable.iterget.get_item_str_uniquified`
    function.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable import iterget

    # Iterable to be tested.
    the_bells = [
        BellType(
            name='Silver bells !',
            stanza='What a world of merriment their melody foretells !'),
        BellType(
            name='Golden bells!',
            stanza='What a world of happiness their harmony foretells !'),
        BellType(
            name='Brazen bells !',
            stanza='What tale of terror, now, their turbulency tells !'),
        BellType(
            name='Iron bells !',
            stanza='What a world of solemn thought their monody compels !'),
    ]

    # Exercise the edge case when this iterable contains no item whose "name"
    # variable matches the passed format specifier.
    the_fifth_bell = iterget.get_item_str_uniquified(
        iterable=the_bells,
        item_attr_name='name',
        item_str_format='Runic bells ! ({})',
    )

    # Assert this getter function to have synthesized the expected string.
    assert the_fifth_bell == 'Runic bells ! (5)'

    # Add a new item whose "name" variable is this string to this iterable.
    the_bells.append(BellType(
        name=the_fifth_bell,
        stanza='In a clamorous appealing to the mercy of the fire,'))

    # Exercise the edge case when this iterable contains one item whose "name"
    # variable matches the passed format specifier.
    the_sixth_bell = iterget.get_item_str_uniquified(
        iterable=the_bells,
        item_attr_name='name',
        item_str_format='Runic bells ! ({})',
    )

    # Assert this getter function to have synthesized the expected string.
    assert the_sixth_bell == 'Runic bells ! (6)'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Project-wide **mathematical operators** (i.e., low-level callables performing
core tensor operations *not* already performed out-of-the-box by either NumPy or
SciPy themselves).
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.typehints import (
    NDArrayNdim1,
    NDArrayNdim1Size2,
    NDArrayNdim2,
)
from numbers import Number

# ....................{ OPERATIONS ~ 1d                    }....................
def det1d(arr1: NDArrayNdim1Size2, arr2: NDArrayNdim1Size2) -> Number:
    '''
    Scalar determinant of two one-dimensional input **2-arrays** (i.e., arrays
    containing exactly two numbers each).

    This function literally computes the following one-liner:

    .. code-block:: pycon

       >>> import numpy as np
       >>> arr1 = np.asarray([1, 2])
       >>> arr2 = np.asarray([4, 5])
       >>> det1d = arr1[0] * arr2[1] - arr1[1] * arr2[0]
       >>> det1d
       -3

    Caveats
    -------
    This function is often (but *not* always) called in lieu of the official
    :func:`numpy.linalg.det` function, which requires that both of the passed
    2-arrays be congealed into a single two-dimensional input array. Weird!

    Parameters
    ----------
    arr1: NDArrayNdim2
        First 1-dimensional input 2-array to take the determinant of.
    arr2: NDArrayNdim2
        Second 1-dimensional input 2-array to take the determinant of.

    Returns
    -------
    Number
        Scalar determinant of these one-dimensional input 2-arrays.
    '''

    # Return us up the one-liner for great justice!
    return arr1[0] * arr2[1] - arr1[1] * arr2[0]

# ....................{ OPERATIONS ~ 2d                    }....................
def cross2d(arr1: NDArrayNdim2, arr2: NDArrayNdim2) -> NDArrayNdim1:
    '''
    One-dimensional cross product of two two-dimensional input arrays.

    This function literally computes the following one-liner:

    .. code-block:: pycon

       >>> import numpy as np
       >>> arr1 = np.asarray([[1, 2], [1, 2]])
       >>> arr2 = np.asarray([[4, 5], [4, 5]])
       >>> cross2d = arr1[..., 0] * arr2[..., 1] - arr1[..., 1] * arr2[..., 0]
       >>> cross2d
       [3 -3]

    Caveats
    -------
    This function should always be called in lieu of the official
    :func:`numpy.linalg.cross` function, which no longer supports this
    two-dimensional use case for inexplicable and frankly unjustifiable reasons:

        DeprecationWarning: Arrays of 2-dimensional vectors are deprecated.
        Use arrays of 3-dimensional vectors instead. (deprecated in NumPy 2.0)

    Parameters
    ----------
    arr1: NDArrayNdim2
        First 2-dimensional input array to take the cross product of.
    arr2: NDArrayNdim2
        Second 2-dimensional input array to take the cross product of.

    Returns
    -------
    NDArrayNdim1
        One-dimensional cross product of these two-dimensional input arrays.
    '''

    # Return us up the one-liner bomb.
    return arr1[..., 0] * arr2[..., 1] - arr1[..., 1] * arr2[..., 0]



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **Python profiling** (i.e., measuring various metrics pertaining to
Python code, including time and space performance) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseMethodUnimplementedException
from betse.util.io.log import logs
from betse.util.type.types import (
    type_check,
    CallableTypes,
    MappingType,
    MappingOrNoneTypes,
    SequenceTypes,
    SequenceOrNoneTypes,
)
from cProfile import Profile
from enum import Enum
from functools import partial
from io import StringIO
from pstats import Stats
from timeit import Timer

# ....................{ ENUMS                             }....................
ProfileType = Enum('ProfileType', ('NONE', 'CALL', 'SIZE',))
# ProfileType = Enum('ProfileType', ('NONE', 'CALL', 'LINE', 'SIZE',))
'''
Enumeration of all possible types of profiling supported by the
:func:`profile_callable` function.

Attributes
----------
NONE : enum
    Null profiling. Callables profiled under this type of profiling will still
    be called as expected but will *not* be profiled.
CALL : enum
    Call-granularity profiling of callables (e.g., functions, lambdas, methods,
    and `eval` statements). This level of granularity is more coarse-grained
    than that of the ``LINE`` type.
SIZE : enum
    Memory profiling of top-level objects returned by top-level callables,
    typically instances of the
    :class:`betse.science.phase.simphaser.SimPhase` class returned by public
    methods of the :class:`betse.science.simrunner.SimRunner` class. This
    profiling type requires installation of the optional third-party dependency
    :mod:`pympler`.
'''
# LINE : enum
#     Line-granularity profiling of callables (e.g., functions, lambdas, methods,
#     and `eval` statements). This level of granularity is more fine-grained than
#     that of the ``CALL`` type, but requires installation of the optional
#     third-party dependency :mod:`line_profiler`.

# ....................{ TIMERS                            }....................
@type_check
def time_callable(
    call: CallableTypes,
    args: SequenceOrNoneTypes = None,
    kwargs: MappingOrNoneTypes = None,
    repetitions: int = 7,
    iterations: int = 1000,
) -> float:
    '''
    Time the passed callable with the passed positional and keyword arguments
    (if any) called the passed number of repetitions of the passed number of
    iterations, returning the time in seconds consumed by the fastest call to
    this callable.

    This function *only* returns the minimum timing of all observed timings,
    discarding the time consumed by all calls other than the fastest call to
    this callable. Why? Because the timing of the fastest call is the most
    statistically signifant timing. All error in timing is **positive,**
    typically due to overhead associated with low-level kernel operations
    (e.g., I/O) and unrelated running processes. By definition, error in timing
    cannot be negative. As astutely noted by the Stackoverflow answer below:

        There's no way to get negative error because a computer can't ever
        compute faster than it can compute!

    Since all timing error is positive, the minimum timing is the timing that
    exhibits minimum error. For all intents and purposes, all other timings are
    effectively irrelevant.

    Parameters
    ----------
    call : CallableTypes
        Callable to be timed.
    args : SequenceOrNoneTypes
        Sequence of positional arguments to pass to each call to this callable.
        Defaults to ``None``, in which no such arguments are passed.
    kwargs: MappingOrNoneTypes
        Dictionary of keyword arguments to pass to each call to this callable.
        Defaults to ``None``, in which no such arguments are passed.
    repetitions : optional[int]
        Number of times to repeat each number of times to call this callable.
        Defaults to a reasonably small value.
    iterations: optional[int]
        Number of times to call this callable for each repetition. The total
        number of calls is thus given by ``repetitions * iterations``. Defaults
        to a reasonably large value.

    See Also
    ----------
    https://stackoverflow.com/a/24105845/2809027
        Stackoverflow answer strongly inspiring this implementation.
    '''

    # Partial function binding this callable to these arguments. Avoid
    # defaulting unpassed positional and keyword arguments to empty data
    # structures, as doing so appears to substantially skew timings and hence
    # should be avoided if feasible.
    callable_bound = None

    # If both positional and keyword arguments were passed, bind this callable
    # to both.
    if args and kwargs:
        callable_bound = partial(call, *args, **kwargs)
    # Else if only positional arguments were passed, bind this callable to only
    # positional arguments.
    elif args:
        callable_bound = partial(call, *args)
    # Else if only keyword arguments were passed, bind this callable to only
    # keyword arguments.
    elif kwargs:
        callable_bound = partial(call, **kwargs)
    # Else, call this callable as is.
    else:
        callable_bound = call

    # Return the minimum timing of this callable repeated this number of times.
    return min(Timer(callable_bound).repeat(repetitions, iterations))

# ....................{ PROFILERS                         }....................
@type_check
def profile_callable(
    call: CallableTypes,
    args: SequenceTypes = None,
    kwargs: MappingType = None,
    is_profile_logged: bool = True,
    profile_filename: str = None,
    profile_type: ProfileType = ProfileType.CALL,
) -> object:
    '''
    Profile the passed callable with the passed positional and keyword
    arguments (if any), returning the value returned by this call and
    optionally logging and serializing the resulting profile to the file with
    the passed filename.

    Parameters
    ----------
    call : CallableTypes
        Callable to be profiled.
    args : optional[SequenceTypes]
        Sequence of positional arguments to pass to the call to this callable.
        Defaults to `None`, in which no such arguments are passed.
    kwargs: optional[MappingType]
        Dictionary of keyword arguments to pass to the call to this callable.
        Defaults to `None`, in which no such arguments are passed.
    is_profile_logged : optional[bool]
        `True` only if logging the profile of this call *after* this call.
        Defaults to `True`.
    profile_filename : optional[str]
        Absolute or relative path of the file to serialize this profile to. If
        this file already exists, this file will be silently overwritten
        *without* explicit warning or error. (While silently overwriting is
        typically discouraged, doing so is sensible in this case. Profiles are
        frequently and trivially created, rendering their preservation
        insignificant by compare to usability concerns.)
        Defaults to `None`, in which case no such file is serialized.
    profile_type : optional[ProfileType]
        Type of profiling to perform. Defaults to :data:`ProfileType.CALL`, in
        which case call-granularity profiling is performed.

    Returns
    ----------
    object
        Value returned by this call.
    '''

    # Private module function performing this type of profiling. Due to type
    # checking, this type is guaranteed to be a valid key of this dictionary.
    profiler = _PROFILE_TYPE_TO_PROFILER[profile_type]

    # Default unpassed positional and keyword arguments.
    if   args is None:   args = ()
    if kwargs is None: kwargs = {}

    # Profile a call of this callable with this profiler, returning the value
    # that this call returns.
    return profiler(
        call=call,
        args=args,
        kwargs=kwargs,
        is_profile_logged=is_profile_logged,
        profile_filename=profile_filename,
    )

# ....................{ PROFILERS ~ none                  }....................
def _profile_callable_none(
    call, args, kwargs, is_profile_logged, profile_filename) -> object:
    '''
    Call the passed callable with the passed positional and keyword arguments
    *without* profiling this call, returning the value returned by this call.

    See Also
    ----------
    :func:`profile_callable`
        Further details on function signature.
    '''

    return call(*args, **kwargs)

# ....................{ PROFILERS ~ call                  }....................
def _profile_callable_call(
    call, args, kwargs, is_profile_logged, profile_filename) -> object:
    '''
    Profile the passed callable in a call-oriented deterministic manner with
    the passed positional and keyword arguments (if any), returning the value
    returned by this call and optionally logging and serializing the resulting
    profile to the file with the passed filename.

    See Also
    ----------
    :func:`profile_callable`
        Further details on function signature.
    '''

    # Log this fact.
    logs.log_debug('Call-granularity profiling enabled.')

    # Call-granularity profile of the subsequent call to this callable.
    profile = Profile()

    # Value returned by calling this callable with these arguments, storing a
    # profile of this call into this "profile" object.
    return_value = profile.runcall(call, *args, **kwargs)

    # If the caller requested this profile be logged...
    if is_profile_logged:
        # Maximum number of slowest callables to log, arbitrarily defined to be
        # twice the default number of rows in the average Linux terminal.
        CALLABLES_MAX = 48

        # String buffer describing these slowest callables.
        calls_sorted = StringIO()

        # Statistics harvested from this profile into this string buffer.
        calls = Stats(profile, stream=calls_sorted)

        # For readability, strip the dirnames from all pathnames in statistics
        # output, reducing these pathnames to basenames.
        calls.strip_dirs()

        # Sort all profiled callables by cumulative time (i.e., total time
        # spent in a callable including all time spent in calls to callables
        # called by that callable).
        calls.sort_stats('cumtime')

        # Write the slowest sublist of these callables to this string buffer.
        calls.print_stats(CALLABLES_MAX)

        # Sort all profiled callables by "total" time (i.e., total time spent
        # in a callable excluding all time spent in calls to callables called
        # by that callable).
        calls.sort_stats('tottime')

        # Write the slowest sublist of these callables to this string buffer.
        calls.print_stats(CALLABLES_MAX)

        # Log this string buffer.
        logs.log_info(
            'Slowest %d callables profiled by both '
            'cumulative and total time:\n%s',
            CALLABLES_MAX, calls_sorted.getvalue())

    # If the caller requested this profile be serialized to a file...
    if profile_filename is not None:
        # Log this serialization.
        logs.log_info(
            'Writing Python-formatted profile to "%s".', profile_filename)

        # Serialize this profile to this file.
        profile.dump_stats(profile_filename)

    # Return the value returned by this call.
    return return_value

# ....................{ PROFILERS ~ line                  }....................
#FIXME: Implement this function properly. Sadly, Python does *NOT* provide an
#out-of-the-box solution for line-based profiling. To do so, either (...both?)
#of the following third-party packages will need to be dynamically detected,
#imported, and leveraged:
#
#* "line", profiling in a line- rather than call-based manner.
#  * "line_profiler", a third-party C extension profiling each line (rather
#    than function as cProfile does). Infrequently updated, but slightly more
#    frequently than "statprof", which is effectively dead. Sadly, the
#    "line_profiler" API is less than ideal, requiring that each individual
#    functions to be non-transitively line-profiled be explicitly decorated
#    with a "line_profiler"-specific decorator -- which, frankly, is crazy and
#    probably not reasonably supportable. See the live repository at:
#    https://github.com/rkern/line_profiler
#  * "statprof", a third-party C extension operating rather differently than
#    either "lineprof" or cProfile. Rather than deterministically instrumenting
#    each line or function call (respectively), "statprof" non-deterministically
#    wakes up at predefined intervals, records a stack trace, and then goes back
#    to sleep. On application completion, "statprof" then tallies up each stack
#    trace and outputs a command-line table of the most expensive lines. Pretty
#    sweet idea. Unsurprisingly, it also appears to be the fastest profiler.
#    Sadly, "statprof" was last updated in 2015; see the dead repository at:
#    https://github.com/bos/statprof.py
#
#Note that the "pprofile" package *CANNOT* be supported, due to its
#encumberment under a GPL-2 license.
#
#In short, there currently appears to be no BSD-compatible, well-maintained
#line profiling solution for Python. Hence, this function is disabled.

def _profile_callable_line(
    call, args, kwargs, is_profile_logged, profile_filename) -> object:
    '''
    Profile the passed callable in a line-oriented deterministic manner with
    the passed positional and keyword arguments (if any), returning the value
    returned by this call and optionally logging and serializing the resulting
    profile to the file with the passed filename.

    See Also
    ----------
    :func:`profile_callable`
        Further details on function signature.
    '''

    # Log this fact.
    logs.log_debug('Line-granularity profiling enabled.')

    # Raise an exception unless this optional dependency is available.
    # libs.die_unless_runtime_optional('????')

    #FIXME: Excise this after properly implementing this function.
    raise BetseMethodUnimplementedException()

# ....................{ PROFILERS ~ size                  }....................
def _profile_callable_size(
    call, args, kwargs, is_profile_logged, profile_filename) -> object:
    '''
    Profile the space consumed by the value returned by calling the passed
    callable with the passed positional and keyword arguments (if any) and
    return this value.

    See Also
    ----------
    :func:`profile_callable`
        Further details on function signature.
    '''

    # Avoid circular import dependencies.
    from betse.lib import libs
    from betse.util.type.obj import objsize

    # Log this fact.
    logs.log_debug('Enabling space (i.e., memory) profiling...')

    # Raise an exception unless the optional "pympler" dependency is available.
    libs.die_unless_runtime_optional('pympler')

    # Value returned by calling this callable with these arguments.
    return_value = call(*args, **kwargs)

    # Log the profiling to be subsequently performed. Since doing so can
    # recurse through the full object tree in the worst case, inform the end
    # user of this potentially slow operation *BEFORE* doing so.
    logs.log_debug(
        'Profiling space (i.e., memory) consumption '
        'for %r() return value of type %r...',
        call, type(return_value))

    # Human-readable string synopsizing this object's memory consumption.
    size_profile = objsize.get_size_profile(
        obj=return_value,

        # Maximum depth of the recursion tree induced by calling this function.
        # To avoid combinatorial explosion, this maximum is necessarily capped
        # to a small positive integer.
        vars_depth=2,

        # Maximum number of the largest instance variables of the value
        # returned by calling this callable to log, arbitrarily defined to be
        # twice the default number of rows in the average Linux terminal.
        vars_max=48,
    )

    # Log this synopsis.
    logs.log_info(
        'Largest top-level object variables '
        'profiled by total space consumption:\n\n'
        'Top-level object %s', size_profile)

    #FIXME: Implement support for serializing this profile to disk.
    # # If the caller requested this profile be serialized to a file...
    # if profile_filename is not None:
    #     # Notify the caller that doing so is currently unsupported..
    #     logs.log_warning(
    #         'Memory profile not saved to "%s", '
    #         'as this feature is currently unimplemented.',
    #         profile_filename)

    # Return the value returned by this call.
    return return_value

# ....................{ GLOBALS ~ private                 }....................
# Technically, the same effect as that of the following dictionary is also
# achievable via getattr() on the current module object. Doing so is
# complicated by artificial constraints Python imposes on doing so (e.g.,
# obtaining the current module object is obscure) and the PEP 20 doctrine of
# "Explicit is better than implicit." We beg to disagree. Nonetheless, the
# explicit approach remains preferable in this edge-case.
_PROFILE_TYPE_TO_PROFILER = {
    ProfileType.CALL: _profile_callable_call,
    # ProfileType.LINE: _profile_callable_line,
    ProfileType.NONE: _profile_callable_none,
    ProfileType.SIZE: _profile_callable_size,
}
'''
Dictionary mapping from each supported type of profiling to the module function
performing this type of profiling.

This private global is intended for use _only_ by the public
:func:`profile_callable` function.
'''
